import numpy as np
import pylab
import tensorflow as tf
from tensorflow import losses, optimizers, nn, keras
from tensorflow.keras import Input, Model, models, layers, metrics, initializers
import os
import skimage
from skimage import transform
import matplotlib.pyplot as plt
from tensorflow.keras.utils import plot_model, model_to_dot
import h5py
from tensorflow.keras.datasets.mnist import load_data
import math


def generator(gen_input_size, relu_before_bn=True):
    trans2d = lambda x, filters: layers.Conv2DTranspose(filters, (5, 5), strides=(2, 2), padding='same', kernel_initializer=initializers.RandomNormal(stddev=0.02))(x)
    dense = lambda x, nodes: layers.Dense(nodes, kernel_initializer=initializers.RandomNormal(stddev=0.02))(x)
    reshape = lambda x, shape: layers.Reshape(shape)(x)
    relu = lambda x: layers.ReLU()(x)
    norm = lambda x: layers.BatchNormalization()(x)
    if relu_before_bn:
        block = lambda x, filters: norm(relu(trans2d(x, filters)))
    else:
        block = lambda x, filters: relu(norm(trans2d(x, filters)))

    inputs = Input(shape=gen_input_size)
    l1 = reshape(dense(inputs, 4 * 4 * 1024), (4, 4, 1024))
    l2 = block(l1, 512)
    l3 = block(l2, 256)
    l4 = block(l3, 128)
    g_out = layers.Conv2DTranspose(3, (5, 5), strides=(2, 2), padding='same', activation='tanh')(l4)

    g_model = Model(inputs=inputs, outputs=g_out)
    return g_model


def discriminator(im_shape, relu_before_bn=True):
    conv2d = lambda x, filters: layers.Conv2D(filters, (5, 5), strides=(2, 2), padding='same', kernel_initializer=initializers.RandomNormal(stddev=0.02))(x)
    flatten = lambda x: layers.Flatten()(x)
    lrelu = lambda x: layers.LeakyReLU(alpha=0.2)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    if relu_before_bn:
        block = lambda x, filters: norm(lrelu(conv2d(x, filters)))
    else:
        block = lambda x, filters: lrelu(norm(conv2d(x, filters)))

    inputs = Input(shape=im_shape)
    l1 = block(inputs, 128)
    l2 = block(l1, 256)
    l3 = block(l2, 512)
    l4 = flatten(block(l3, 1024))
    d_out = layers.Dense(1, activation='sigmoid')(l4)

    d_model = Model(inputs=inputs, outputs=d_out)
    opt = optimizers.Adam(learning_rate=0.0002, beta_1=0.5)
    d_model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return d_model


def gan(g_model, d_model):
    d_model.trainable = False
    d_output = d_model(g_model.output)

    gan_model = Model(inputs=g_model.input, outputs=d_output)
    opt = optimizers.Adam(learning_rate=0.0002, beta_1=0.5)
    gan_model.compile(loss='binary_crossentropy', optimizer=opt)
    return gan_model


def load_real_samples():
    (trainX, _), (_, _) = load_data()
    X = np.expand_dims(trainX, axis=-1)
    X = X.astype('float32')
    X = (X-127.5) / 127.5
    return X


def generate_real_samples(dataset, n_samples):
    ix = np.random.randint(0, dataset.shape[0], n_samples)
    X = dataset[ix]
    y = np.ones((n_samples, 1))
    return X, y


def generate_latent_points(latent_dim, n_samples):
    x_input = np.random.randn(latent_dim * n_samples)
    x_input = x_input.reshape(n_samples, latent_dim)
    return x_input


def generate_fake_samples(g_model, latent_dim, n_samples):
    x_input = generate_latent_points(latent_dim, n_samples)
    X = g_model.predict(x_input)
    y = np.zeros((n_samples, 1))
    return X, y


def save_plot(examples, epoch, n=10):
    for i in range(n * n):
        plt.subplot(n, n, 1 + i)
        plt.axis('off')
        plt.imshow(examples[i, :, :, :])
    filename = 'WSI_DCGAN_generated_plot_e%03d.pdf' % (epoch + 1)
    plt.savefig(filename)
    plt.close()


def summarize_performance(epoch, g_model, d_model, dataset, latent_dim, n_samples=100):
    X_real, y_real = generate_real_samples(dataset, n_samples)
    _, acc_real = d_model.evaluate(X_real, y_real, verbose=0)
    x_fake, y_fake = generate_fake_samples(g_model, latent_dim, n_samples)
    _, acc_fake = d_model.evaluate(x_fake, y_fake, verbose=0)
    print('>Accuracy real: %.0f%%, fake: %.0f%%' % (acc_real * 100, acc_fake * 100))
    save_plot(x_fake, epoch)
    filename = 'WSI_DCGAN_generator_model_%03d.h5' % (epoch + 1)
    g_model.save(filename)


def train(g_model, d_model, gan_model, dataset, latent_dim, n_epochs=100, n_batch=256):
    bat_per_epo = int(dataset.shape[0] / n_batch)
    half_batch = int(n_batch / 2)
    for i in range(n_epochs):
        for j in range(bat_per_epo):
            X_real, y_real = generate_real_samples(dataset, half_batch)
            X_fake, y_fake = generate_fake_samples(g_model, latent_dim, half_batch)
            # need to separate real and fake minibatches
            X, y = np.vstack((X_real, X_fake)), np.vstack((y_real, y_fake))
            # note that the underscore will cause problems if d_model is compiles without accuracy metric
            d_loss, _ = d_model.train_on_batch(X, y)
            X_gan = generate_latent_points(latent_dim, n_batch)
            y_gan = np.ones((n_batch, 1))
            g_loss = gan_model.train_on_batch(X_gan, y_gan)
            print('>%d, %d/%d, d=%.3f, g=%.3f' % (i + 1, j + 1, bat_per_epo, d_loss, g_loss))
        if (i + 1) % 5 == 0:
            summarize_performance(i, g_model, d_model, dataset, latent_dim)



import numpy as np
from tensorflow.keras import Input, Model, layers
from jarvis.utils.display import imshow



def retinanet_resnet50_3d(inputs, K, A, filter_ratio=1, n=2, include_fc_layer=False, shared_weights=False):
    """Generates retinanet with resnet backbone. Can specify if classification and regression networks share weights"""
    r_model = resnet50_3d(inputs=inputs['dat'],
                        filter_ratio=filter_ratio,
                        n=n,
                        include_fc_layer=include_fc_layer)
    backbone_output = [r_model.get_layer(layer_name).output for layer_name in ['c3-output', 'c4-output', 'c5-output']]
    fp_out = feature_pyramid_3d(inputs=backbone_output,
                                filter_ratio=filter_ratio)
    logits = class_and_reg_subnets(feature_pyramid=fp_out,
                                   K=K,
                                   A=A,
                                   filter_ratio=filter_ratio,
                                   shared_weights=shared_weights)
    model = Model(inputs=inputs, outputs=logits)
    return model


def resnet50_3d(inputs, filter_ratio=1, n=2, include_fc_layer=False):
    """

    :param inputs: Keras Input object with desire shape
    :type inputs:
    :param filter_ratio:
    :type filter_ratio:
    :param n: # of categories
    :type n: integer
    :param include_fc_layer:
    :type include_fc_layer:
    :return:
    :rtype:
    """
    # --- Define kwargs dictionary
    kwargs1 = {
        'kernel_size': (1, 1, 1),
        'padding': 'valid',
    }
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    kwargs7 = {
        'kernel_size': (1, 7, 7),
        'padding': 'valid',
    }
    # --- Define block components
    conv1 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs1)(x)
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    conv7 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs7)(x)
    max_pool = lambda x, pool_size, strides: layers.MaxPooling3D(pool_size=pool_size,
                                                                 strides=strides,
                                                                 padding='valid')(x)
    norm = lambda x: layers.BatchNormalization()(x)
    add = lambda x, y: layers.Add()([x, y])
    zeropad = lambda x, padding: layers.ZeroPadding3D(padding=padding)(x)
    # --- Residual blocks
    # conv blocks
    conv_1 = lambda filters, x, strides: relu(norm(conv1(x,
                                                         filters,
                                                         strides=strides)))
    conv_2 = lambda filters, x: relu(norm(conv3(x,
                                                filters,
                                                strides=1)))
    conv_3 = lambda filters, x: norm(conv1(x,
                                           filters,
                                           strides=1))
    conv_sc = lambda filters, x, strides: norm(conv1(x,
                                                     filters,
                                                     strides=strides))
    conv_block = lambda filters1, filters2, x, strides: relu(add(conv_sc(filters2,
                                                                         x,
                                                                         strides),
                                                                 conv_3(filters2,
                                                                        conv_2(filters1,
                                                                               conv_1(filters1,
                                                                                      x,
                                                                                      strides)))))
    # identity blocks
    identity_1 = lambda filters, x: relu(norm(conv1(x,
                                                    filters,
                                                    strides=1)))
    identity_2 = lambda filters, x: relu(norm(conv3(x,
                                                    filters,
                                                    strides=1)))
    identity_3 = lambda filters, x: norm(conv1(x,
                                               filters,
                                               strides=1))
    identity_block = lambda filters1, filters2, x: relu(add(identity_3(filters2,
                                                                       identity_2(filters1,
                                                                                  identity_1(filters1,
                                                                                             x))),
                                                            x))
    # --- ResNet-50 backbone
    # stage 1 c2 1/4
    res1 = max_pool(zeropad(relu(norm(conv7(zeropad(inputs,
                                                    (0, 3, 3)),
                                            int(64*filter_ratio),
                                            strides=(1, 2, 2)))),
                            (0, 1, 1)),
                    (1, 3, 3),
                    strides=(1, 2, 2))
    # stage 2 c2 1/4
    res2 = layers.Lambda(lambda x: x, name='c2-output')(
        identity_block(int(64*filter_ratio),
                       int(256*filter_ratio),
                       identity_block(int(64*filter_ratio),
                                      int(256*filter_ratio),
                                      conv_block(int(64*filter_ratio),
                                                 int(256*filter_ratio),
                                                 res1,
                                                 strides=1)))
    )
    # stage 3 c3 1/8
    res3 = layers.Lambda(lambda x: x, name='c3-output')(
        identity_block(int(128*filter_ratio),
                       int(512*filter_ratio),
                       identity_block(int(128*filter_ratio),
                                      int(512*filter_ratio),
                                      identity_block(int(128*filter_ratio),
                                                     int(512*filter_ratio),
                                                     conv_block(int(128*filter_ratio),
                                                                int(512*filter_ratio),
                                                                res2,
                                                                strides=(1, 2, 2)))))
    )
    # stage 4 c4 1/16
    res4 = layers.Lambda(lambda x: x, name='c4-output')(
        identity_block(int(256*filter_ratio),
                       int(1024*filter_ratio),
                       identity_block(int(256*filter_ratio),
                                      int(1024*filter_ratio),
                                      identity_block(int(256*filter_ratio),
                                                     int(1024*filter_ratio),
                                                     identity_block(int(256*filter_ratio),
                                                                    int(1024*filter_ratio),
                                                                    identity_block(int(256*filter_ratio),
                                                                                   int(1024*filter_ratio),
                                                                                   conv_block(int(256*filter_ratio),
                                                                                              int(1024*filter_ratio),
                                                                                              res3,
                                                                                              strides=(1, 2, 2)))))))
    )
    # stage 5 c5 1/32
    res5 = layers.Lambda(lambda x: x, name='c5-output')(
        identity_block(int(512*filter_ratio),
                       int(2048*filter_ratio),
                       identity_block(int(512*filter_ratio),
                                      int(2048*filter_ratio),
                                      conv_block(int(512*filter_ratio),
                                                 int(2048*filter_ratio),
                                                 res4,
                                                 strides=(1, 2, 2))))
    )
    avg_pool = layers.GlobalAveragePooling3D()(res5)
    flatten = layers.Flatten()(avg_pool)
    logits = layers.Dense(n)(flatten)
    if include_fc_layer:
        model = Model(inputs=inputs, outputs=logits)
    else:
        model = Model(inputs=inputs, outputs=res5)
    return model


def feature_pyramid_3d(inputs, filter_ratio):
    kwargs1 = {
        'kernel_size': (1, 1, 1),
        'padding': 'valid',
    }
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    conv1 = lambda x, filters, strides: layers.Conv3D(filters=filters, strides=strides, **kwargs1)(x)
    add = lambda x, y: layers.Add()([x, y])
    upsamp2x = lambda x: layers.UpSampling3D(size=(1, 2, 2))(x)
    fp_block = lambda x, y: add(upsamp2x(x), conv1(y, int(256*filter_ratio), strides=1))
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters, strides=strides, **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)

    p5 = conv1(inputs[2], int(256*filter_ratio), strides=1)
    fp4 = fp_block(p5, inputs[1])
    p4 = conv3(fp4, int(256*filter_ratio), strides=1)
    fp3 = fp_block(fp4, inputs[0])
    p3 = conv3(fp3, int(256*filter_ratio), strides=1)
    p6 = conv3(p5, int(256*filter_ratio), strides=(1, 2, 2))
    p7 = conv3(relu(p6), int(256*filter_ratio), strides=(1, 2, 2))
    return [p3, p4, p5, p6, p7]


def class_and_reg_subnets(feature_pyramid, K, A, filter_ratio, shared_weights=False):
    if shared_weights:
        class_subnet = classification_head(K, A, filter_ratio)
        box_subnet = regression_head(A, filter_ratio)
        class_outputs = [class_subnet(features) for features in feature_pyramid]
        box_outputs = [box_subnet(features) for features in feature_pyramid]
        logits = {'cls-c3': layers.Lambda(lambda x: x, name='cls-c3')(class_outputs[0]),
                  'reg-c3': layers.Lambda(lambda x: x, name='reg-c3')(box_outputs[0]),
                  'cls-c4': layers.Lambda(lambda x: x, name='cls-c4')(class_outputs[1]),
                  'reg-c4': layers.Lambda(lambda x: x, name='reg-c4')(box_outputs[1]),
                  'cls-c5': layers.Lambda(lambda x: x, name='cls-c5')(class_outputs[2]),
                  'reg-c5': layers.Lambda(lambda x: x, name='reg-c5')(box_outputs[2])}
        return logits
    else:
        class_models = []
        reg_models = []
        for _ in feature_pyramid:
            class_models.append(classification_head(K, A, filter_ratio))
            reg_models.append(regression_head(A, filter_ratio))
        logits = {'cls-c3': layers.Lambda(lambda x: x, name='cls-c3')(class_models[0](feature_pyramid[0])),
                  'reg-c3': layers.Lambda(lambda x: x, name='reg-c3')(reg_models[0](feature_pyramid[0])),
                  'cls-c4': layers.Lambda(lambda x: x, name='cls-c4')(class_models[1](feature_pyramid[1])),
                  'reg-c4': layers.Lambda(lambda x: x, name='reg-c4')(reg_models[1](feature_pyramid[1])),
                  'cls-c5': layers.Lambda(lambda x: x, name='cls-c5')(class_models[2](feature_pyramid[2])),
                  'reg-c5': layers.Lambda(lambda x: x, name='reg-c5')(reg_models[2](feature_pyramid[2]))}
        return logits


def retinanet_resnet_2(inputs, K, A):
    """Retinanet with resnet backbone. Classification and regression networks have different weights for each feature
    pyramid layer"""
    # --- Define kwargs dictionary
    kwargs1 = {
        'kernel_size': (1, 1, 1),
        'padding': 'valid',
    }
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    kwargs7 = {
        'kernel_size': (1, 7, 7),
        'padding': 'valid',
    }
    # --- Define block components
    conv1 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs1)(x)
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    conv7 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs7)(x)
    max_pool = lambda x, pool_size, strides: layers.MaxPooling3D(pool_size=pool_size,
                                                                 strides=strides,
                                                                 padding='valid')(x)
    norm = lambda x: layers.BatchNormalization()(x)
    add = lambda x, y: layers.Add()([x, y])
    zeropad = lambda x, padding: layers.ZeroPadding3D(padding=padding)(x)
    upsamp2x = lambda x: layers.UpSampling3D(size=(1, 2, 2))(x)
    # --- Define stride-1, stride-2 blocks
    # conv1 = lambda filters, x : relu(conv(x, filters, strides=1))
    # conv2 = lambda filters, x : relu(conv(x, filters, strides=(2, 2)))
    # --- Residual blocks
    # conv blocks
    conv_1 = lambda filters, x, strides: relu(norm(conv1(x,
                                                         filters,
                                                         strides=strides)))
    conv_2 = lambda filters, x: relu(norm(conv3(x,
                                                filters,
                                                strides=1)))
    conv_3 = lambda filters, x: norm(conv1(x,
                                           filters,
                                           strides=1))
    conv_sc = lambda filters, x, strides: norm(conv1(x,
                                                     filters,
                                                     strides=strides))
    conv_block = lambda filters1, filters2, x, strides: relu(add(conv_3(filters2,
                                                                        conv_2(filters1,
                                                                               conv_1(filters1,
                                                                                      x,
                                                                                      strides))),
                                                                 conv_sc(filters2,
                                                                         x,
                                                                         strides)))
    # identity blocks
    identity_1 = lambda filters, x: relu(norm(conv1(x,
                                                    filters,
                                                    strides=1)))
    identity_2 = lambda filters, x: relu(norm(conv3(x,
                                                    filters,
                                                    strides=1)))
    identity_3 = lambda filters, x: norm(conv1(x,
                                               filters,
                                               strides=1))
    identity_block = lambda filters1, filters2, x: relu(add(identity_3(filters2,
                                                                       identity_2(filters1,
                                                                                  identity_1(filters1,
                                                                                             x))),
                                                            x))
    # --- feature pyramid blocks
    fp_block = lambda x, y: add(upsamp2x(x), conv1(y,
                                                   256,
                                                   strides=1))
    # --- class subnet blocks
    conv_class1 = lambda filters, x: relu(conv3(x,
                                                filters,
                                                strides=1))
    # --- classification head
    class_subnet3 = classification_head(K, A)
    class_subnet4 = classification_head(K, A)
    class_subnet5 = classification_head(K, A)
    # --- regression head
    box_subnet3 = regression_head(A)
    box_subnet4 = regression_head(A)
    box_subnet5 = regression_head(A)
    # --- ResNet-50 backbone
    # stage 1 c2 1/4
    res1 = max_pool(zeropad(relu(norm(conv7(zeropad(inputs['dat'],
                                                    (0, 3, 3)),
                                            64,
                                            strides=(1, 2, 2)))),
                            (0, 1, 1)),
                    (1, 3, 3),
                    strides=(1, 2, 2))
    # stage 2 c2 1/4
    res2 = identity_block(64,
                          256,
                          identity_block(64,
                                         256,
                                         conv_block(64,
                                                    256,
                                                    res1,
                                                    strides=1)))
    # stage 3 c3 1/8
    res3 = identity_block(128,
                          512,
                          identity_block(128,
                                         512,
                                         identity_block(128,
                                                        512,
                                                        conv_block(128,
                                                                   512,
                                                                   res2,
                                                                   strides=(1, 2, 2)))))
    # stage 4 c4 1/16
    res4 = identity_block(256,
                          1024,
                          identity_block(256,
                                         1024,
                                         identity_block(256,
                                                        1024,
                                                        identity_block(256,
                                                                       1024,
                                                                       identity_block(256,
                                                                                      1024,
                                                                                      conv_block(256,
                                                                                                 1024,
                                                                                                 res3,
                                                                                                 strides=(1, 2, 2)))))))
    # stage 5 c5 1/32
    res5 = identity_block(512,
                          2048,
                          identity_block(512,
                                         2048,
                                         conv_block(512,
                                                    2048,
                                                    res4,
                                                    strides=(1, 2, 2))))
    # --- Feature Pyramid Network architecture
    # p5 1/32
    fp5 = conv1(res5, 256, strides=1)
    # p4 1/16
    fp4 = fp_block(fp5, res4)
    p4 = conv3(fp4, 256, strides=1)
    # p3 1/8
    fp3 = fp_block(fp4, res3)
    p3 = conv3(fp3, 256, strides=1)
    # p6 1/4
    # p6 = conv3(fp5, 256, strides=(2, 2))
    # p7 1/2
    # p7 = conv3(relu(p6), 256, strides=(2, 2))
    # feature_pyramid = [p3, p4, fp5]
    # lambda layer that allows multiple outputs from a shared model to have specific names
    # layers.Lambda(lambda x:x, name=name)()
    # --- Class subnet
    # class_outputs = [class_subnet(features) for features in feature_pyramid]
    # --- Box subnet
    # box_outputs = [box_subnet(features) for features in feature_pyramid]
    # --- put class and box outputs in dictionary
    logits = {'cls-c3': layers.Lambda(lambda x: x, name='cls-c3')(class_subnet3(p3)),
              'reg-c3': layers.Lambda(lambda x: x, name='reg-c3')(box_subnet3(p3)),
              'cls-c4': layers.Lambda(lambda x: x, name='cls-c4')(class_subnet4(p4)),
              'reg-c4': layers.Lambda(lambda x: x, name='reg-c4')(box_subnet4(p4)),
              'cls-c5': layers.Lambda(lambda x: x, name='cls-c5')(class_subnet5(fp5)),
              'reg-c5': layers.Lambda(lambda x: x, name='reg-c5')(box_subnet5(fp5))}

    model = Model(inputs=inputs, outputs=logits)
    return model


def retinanet_resnet(inputs, K, A):
    """Retinanet with resnet backbone. Classification and regression networks share weights across feature pyramid
     layers"""
    # --- Define kwargs dictionary
    kwargs1 = {
        'kernel_size': (1, 1, 1),
        'padding': 'valid',
    }
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    kwargs7 = {
        'kernel_size': (1, 7, 7),
        'padding': 'valid',
    }
    # --- Define block components
    conv1 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs1)(x)
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    conv7 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs7)(x)
    max_pool = lambda x, pool_size, strides: layers.MaxPooling3D(pool_size=pool_size,
                                                                 strides=strides,
                                                                 padding='valid')(x)
    norm = lambda x: layers.BatchNormalization()(x)
    add = lambda x, y: layers.Add()([x, y])
    zeropad = lambda x, padding: layers.ZeroPadding3D(padding=padding)(x)
    upsamp2x = lambda x: layers.UpSampling3D(size=(1, 2, 2))(x)
    # --- Define stride-1, stride-2 blocks
    # conv1 = lambda filters, x : relu(conv(x, filters, strides=1))
    # conv2 = lambda filters, x : relu(conv(x, filters, strides=(2, 2)))
    # --- Residual blocks
    # conv blocks
    conv_1 = lambda filters, x, strides: relu(norm(conv1(x,
                                                         filters,
                                                         strides=strides)))
    conv_2 = lambda filters, x: relu(norm(conv3(x,
                                                filters,
                                                strides=1)))
    conv_3 = lambda filters, x: norm(conv1(x,
                                           filters,
                                           strides=1))
    conv_sc = lambda filters, x, strides: norm(conv1(x,
                                                     filters,
                                                     strides=strides))
    conv_block = lambda filters1, filters2, x, strides: relu(add(conv_3(filters2,
                                                                        conv_2(filters1,
                                                                               conv_1(filters1,
                                                                                      x,
                                                                                      strides))),
                                                                 conv_sc(filters2,
                                                                         x,
                                                                         strides)))
    # identity blocks
    identity_1 = lambda filters, x: relu(norm(conv1(x,
                                                    filters,
                                                    strides=1)))
    identity_2 = lambda filters, x: relu(norm(conv3(x,
                                                    filters,
                                                    strides=1)))
    identity_3 = lambda filters, x: norm(conv1(x,
                                               filters,
                                               strides=1))
    identity_block = lambda filters1, filters2, x: relu(add(identity_3(filters2,
                                                                       identity_2(filters1,
                                                                                  identity_1(filters1,
                                                                                             x))),
                                                            x))
    # --- feature pyramid blocks
    fp_block = lambda x, y: add(upsamp2x(x), conv1(y, 256, strides=1))
    # --- classification head
    class_subnet = classification_head(K, A)
    # --- regression head
    box_subnet = regression_head(A)
    # --- ResNet-50 backbone
    # stage 1 c2 1/4
    res1 = max_pool(zeropad(relu(norm(conv7(zeropad(inputs['dat'],
                                                    (0, 3, 3)),
                                            64,
                                            strides=(1, 2, 2)))),
                            (0, 1, 1)),
                    (1, 3, 3),
                    strides=(1, 2, 2))
    # stage 2 c2 1/4
    res2 = identity_block(64,
                          256,
                          identity_block(64,
                                         256,
                                         conv_block(64,
                                                    256,
                                                    res1,
                                                    strides=1)))
    # stage 3 c3 1/8
    res3 = identity_block(128,
                          512,
                          identity_block(128,
                                         512,
                                         identity_block(128,
                                                        512,
                                                        conv_block(128,
                                                                   512,
                                                                   res2,
                                                                   strides=(1, 2, 2)))))
    # stage 4 c4 1/16
    res4 = identity_block(256,
                          1024,
                          identity_block(256,
                                         1024,
                                         identity_block(256,
                                                        1024,
                                                        identity_block(256,
                                                                       1024,
                                                                       identity_block(256,
                                                                                      1024,
                                                                                      conv_block(256,
                                                                                                 1024,
                                                                                                 res3,
                                                                                                 strides=(1, 2, 2)))))))
    # stage 5 c5 1/32
    res5 = identity_block(512,
                          2048,
                          identity_block(512,
                                         2048,
                                         conv_block(512,
                                                    2048,
                                                    res4,
                                                    strides=(1, 2, 2))))
    # --- Feature Pyramid Network architecture
    # p5 1/32
    fp5 = conv1(res5, 256, strides=1)
    # p4 1/16
    fp4 = fp_block(fp5, res4)
    p4 = conv3(fp4, 256, strides=1)
    # p3 1/8
    fp3 = fp_block(fp4, res3)
    p3 = conv3(fp3, 256, strides=1)
    # p6 1/4
    # p6 = conv3(fp5, 256, strides=(2, 2))
    # p7 1/2
    # p7 = conv3(relu(p6), 256, strides=(2, 2))
    feature_pyramid = [p3, p4, fp5]
    # lambda layer that allows multiple outputs from a shared model to have specific names
    # layers.Lambda(lambda x:x, name=name)()
    # --- Class subnet
    class_outputs = [class_subnet(features) for features in feature_pyramid]
    # --- Box subnet
    box_outputs = [box_subnet(features) for features in feature_pyramid]
    # --- put class and box outputs in dictionary
    logits = {'cls-c3': layers.Lambda(lambda x: x, name='cls-c3')(class_outputs[0]),
              'reg-c3': layers.Lambda(lambda x: x, name='reg-c3')(box_outputs[0]),
              'cls-c4': layers.Lambda(lambda x: x, name='cls-c4')(class_outputs[1]),
              'reg-c4': layers.Lambda(lambda x: x, name='reg-c4')(box_outputs[1]),
              'cls-c5': layers.Lambda(lambda x: x, name='cls-c5')(class_outputs[2]),
              'reg-c5': layers.Lambda(lambda x: x, name='reg-c5')(box_outputs[2])}

    model = Model(inputs=inputs, outputs=logits)
    return model


def classification_head(K, A, filter_ratio=1):
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    conv_class1 = lambda filters, x: relu(conv3(x,
                                                filters,
                                                strides=1))

    inputs = Input(shape=(None, None, None, int(256*filter_ratio)))
    c1 = conv_class1(int(256*filter_ratio),
                     conv_class1(int(256*filter_ratio),
                                 conv_class1(int(256*filter_ratio),
                                             conv_class1(int(256*filter_ratio),
                                                         inputs))))
    c2 = conv3(c1, K*A, strides=1)
    class_subnet = Model(inputs=inputs, outputs=c2)
    return class_subnet


def regression_head(A, filter_ratio=1):
    kwargs3 = {
        'kernel_size': (1, 3, 3),
        'padding': 'same',
    }
    conv3 = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                      strides=strides,
                                                      **kwargs3)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    conv_class1 = lambda filters, x: relu(conv3(x,
                                                filters,
                                                strides=1))

    inputs = Input(shape=(None, None, None, int(256*filter_ratio)))
    b1 = conv_class1(int(256*filter_ratio),
                     conv_class1(int(256*filter_ratio),
                                 conv_class1(int(256*filter_ratio),
                                             conv_class1(int(256*filter_ratio),
                                                         inputs))))
    b2 = conv3(b1, 4*A, strides=1)

    box_subnet = Model(inputs=inputs, outputs=b2)
    return box_subnet


def retinanet(inputs, K, A):
    """Retinanet architecture with peter's simple backbone architecture."""
    kwargs = {
        'kernel_size': (1, 3, 3),
        'padding': 'same'
    }
    conv = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                     strides=strides,
                                                     **kwargs)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    relu = lambda x: layers.LeakyReLU()(x)

    conv1 = lambda filters, x: relu(norm(conv(x,
                                              filters,
                                              strides=1)))
    conv2 = lambda filters, x: relu(norm(conv(x,
                                              filters,
                                              strides=(1, 2, 2))))

    l1 = conv1(8, inputs['dat'])
    l2 = conv1(16, conv2(16, l1))
    l3 = conv1(24, conv2(24, l2))
    l4 = conv1(32, conv2(32, l3))
    l5 = conv1(48, conv2(48, l4))
    l6 = conv1(64, conv2(64, l5))

    zoom = lambda x: layers.UpSampling3D(size=(1, 2, 2))(x)

    proj = lambda filters, x: layers.Conv3D(filters=filters,
                                            strides=1,
                                            kernel_size=(1, 1, 1),
                                            padding='same',
                                            kernel_initializer='he_normal')(x)

    l7 = proj(64, l6)
    l8 = conv1(64, zoom(l7) + proj(64, l5))
    l9 = conv1(64, zoom(l8) + proj(64, l4))

    logits = {}
    K = K
    A = A

    # --- C2
    c3_cls = conv1(64, conv1(64, l9))
    c3_reg = conv1(64, conv1(64, l9))
    logits['cls-c3'] = layers.Conv3D(filters=(A * K), name='cls-c3', **kwargs)(c3_cls)
    logits['reg-c3'] = layers.Conv3D(filters=(A * 4), name='reg-c3', **kwargs)(c3_reg)

    # --- C3
    c4_cls = conv1(64, conv1(64, l8))
    c4_reg = conv1(64, conv1(64, l8))
    logits['cls-c4'] = layers.Conv3D(filters=(A * K), name='cls-c4', **kwargs)(c4_cls)
    logits['reg-c4'] = layers.Conv3D(filters=(A * 4), name='reg-c4', **kwargs)(c4_reg)

    # --- C4
    c5_cls = conv1(64, conv1(64, l7))
    c5_reg = conv1(64, conv1(64, l7))
    logits['cls-c5'] = layers.Conv3D(filters=(A * K), name='cls-c5', **kwargs)(c5_cls)
    logits['reg-c5'] = layers.Conv3D(filters=(A * 4), name='reg-c5', **kwargs)(c5_reg)

    model = Model(inputs=inputs, outputs=logits)
    return model


def generator(data, batchsize=1):
    """Data generator for retinanet"""
    i = 0
    keys = data.keys()
    while True:
        if i == (len(data['dat']) // batchsize):
            i = 0
            p = np.random.permutation(len(data['dat']))
            for key in keys:
                data[key] = data[key][p]
        start = i * batchsize
        stop = start + batchsize
        xbatch = {}
        ybatch = {}
        for key in keys:
            if 'dat' in key:
                xbatch[key] = data[key][start:stop]
            elif 'msk' in key:
                xbatch[key] = data[key][start:stop]
            else:
                ybatch[key] = data[key][start:stop]
        i += 1
        yield xbatch, ybatch


def compare_specific_results(data, i, iou_nms, model, boundingbox):
    test_im_x = {}
    test_im_y = {}
    for key in data.keys():
        if 'dat' in key:
            test_im_x[key] = np.expand_dims(data[key][i], axis=0)
        elif 'msk' in key:
            test_im_x[key] = np.expand_dims(data[key][i], axis=0)
        else:
            test_im_y[key] = np.expand_dims(data[key][i], axis=0)
    test1 = model.predict(test_im_x)
    test1_dict = {name: pred for name, pred in zip(model.output_names, test1)}

    # --- Convert box to mask (for visualization)
    msk = boundingbox.convert_box_to_msk(test1_dict, iou_nms=iou_nms, apply_deltas=True)
    imshow(data['dat'][i, :, :, :, 2], msk, title='AI-generated boxes')

    # box = bb.convert_anc_to_box(bbox[300], np.ones((bbox[300].shape[0],1)))
    # --- Convert box to mask (for visualization)
    msk = boundingbox.convert_box_to_msk(test_im_y, apply_deltas=True)
    imshow(data['dat'][i, :, :, :, 2], msk, title='Ground-truth template boxes')
    return test1_dict


def unet(inputs, num_layers=6, num_classes=2, _3d=False):
    # --- Define kwargs dictionary
    if _3d:
        kwargs = {
            'kernel_size': (3, 3, 3),
            'padding': 'same'}
    else:
        kwargs = {
            'kernel_size': (1, 3, 3),
            'padding': 'same'}

    # --- Define lambda functions
    conv = lambda x, filters, strides: layers.Conv3D(filters=filters,
                                                     strides=strides,
                                                     **kwargs)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    relu = lambda x: layers.LeakyReLU()(x)
    tran = lambda x, filters, strides: layers.Conv3DTranspose(filters=filters,
                                                              strides=strides,
                                                              **kwargs)(x)

    # --- Define stride-1, stride-2 blocks
    conv1 = lambda filters, x: relu(norm(conv(x, filters, strides=1)))
    conv2 = lambda filters, x: relu(norm(conv(x, filters, strides=(1, 2, 2))))
    tran2 = lambda filters, x: relu(norm(tran(x, filters, strides=(1, 2, 2))))

    # --- Define simple layers
    c_layer = lambda filters, x: conv1(filters, conv2(filters, x))
    e_layer = lambda filters1, filters2, x: tran2(filters1, conv1(filters2, x))

    contracting_layers = []
    for i in range(num_layers):  # 0,1,2,3,4,5
        if i == 0:
            contracting_layers.append(conv1(8, inputs['dat']))
        else:
            contracting_layers.append(c_layer(16*i, contracting_layers[i-1]))
    expanding_layers = []
    for j in reversed(range(num_layers-1)):  # 4,3,2,1,0
        if j == num_layers-2:
            expanding_layers.append(tran2(16*j, contracting_layers[j+1]))
        else:
            expanding_layers.append(e_layer(16*j if j != 0 else 8,
                                            16*(j+1),
                                            expanding_layers[-1] + contracting_layers[j+1]))
    last_layer = conv1(8, conv1(8, expanding_layers[-1] + contracting_layers[0]))

    # --- Create logits
    logits = {}
    logits['zones'] = layers.Conv3D(filters=num_classes, name='zones', **kwargs)(last_layer)

    # --- Create model
    model = Model(inputs=inputs, outputs=logits)
    return model



import numpy as np
import pylab
import tensorflow as tf
from tensorflow import losses, optimizers, nn, keras
from tensorflow.keras import Input, Model, models, layers, metrics, initializers
import os
import skimage
from skimage import transform
import matplotlib.pyplot as plt
from tensorflow.keras.utils import plot_model, model_to_dot
import h5py
from tensorflow.keras.datasets.mnist import load_data
import math


def encoder(im_shape, relu_before_bn=True):
    conv2d = lambda x, filters: layers.Conv2D(filters, (3, 3), strides=(2, 2), padding='same')(x)
    lrelu = lambda x: layers.LeakyReLU(alpha=0.2)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    if relu_before_bn:
        block = lambda x, filters: norm(lrelu(conv2d(x, filters)))
    else:
        block = lambda x, filters: lrelu(norm(conv2d(x, filters)))

    inputs = Input(shape=im_shape)
    l1 = block(inputs, 32)
    l2 = block(l1, 16)
    l3 = block(l2, 8)
    encoder_out = block(l3, 8)


    encoder_model = Model(inputs=inputs, outputs=encoder_out)
    return encoder_model

def decoder(gen_input_size, relu_before_bn=True):
    trans2d = lambda x, filters: layers.Conv2DTranspose(filters, (3, 3), strides=(2, 2), padding='same')(x)
    lrelu = lambda x: layers.LeakyReLU(alpha=0.2)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    if relu_before_bn:
        block = lambda x, filters: norm(lrelu(trans2d(x, filters)))
    else:
        block = lambda x, filters: lrelu(norm(trans2d(x, filters)))

    inputs = Input(shape=gen_input_size)
    l1 = block(inputs, 8)
    l2 = block(l1, 16)
    l3 = block(l2, 32)
    decoder_out = layers.Conv2DTranspose(3, (3, 3), strides=(2, 2), padding='same', activation='tanh')(l3)

    decoder_model = Model(inputs=inputs, outputs=decoder_out)
    return decoder_model


def autoencoder(encoder_model, decoder_model):
    decoder_output = decoder_model(encoder_model.output)

    model = Model(inputs=encoder_model.input, outputs=decoder_output)
    opt = optimizers.Adam()
    model.compile(loss='mse', optimizer=opt)
    return model


def load_data(filename, tanh=True):
    dataset = np.load(filename)
    dataset.astype('float32')
    if tanh:
        dataset = (dataset-127.5)/127.5
    else:
        dataset = dataset/255
    return dataset



import numpy as np
import matplotlib.pyplot as plt
import scipy
import scipy.stats
import sklearn
from sklearn.neighbors import KDTree
import matplotlib
from mpl_toolkits.mplot3d import Axes3D
import random
import pandas as pd


def agreement_table(a, b, c):
    """
    Returns an agreement table as a 2D array.

    Meant for use in cohens_kappa() function.

    :param a: count values for observer #1
    :type a: ndarray of ints [Yes, No]
    :param b: count values for observer #2
    :type b: ndarray of ints [Yes, No]
    :param c: agreement count values between the 2 observers
    :type c: ndarray of ints [Yes, No]
    :return: agreement table
    :rtype: ndarray
    """
    if (len(a[0]) + len(a[1])) != (len(b[0]) + len(b[1])):
        print('total annotations do not match each other!')
        return
    table = np.empty([3, 3])
    table[0, 0] = c[0]
    table[0, 1] = a[1] - c[1]
    table[1, 0] = a[0] - c[0]
    table[1, 1] = c[1]
    table[0, 2] = b[0]
    table[1, 2] = b[1]
    table[2, 0] = a[0]
    table[2, 1] = a[1]
    table[2, 2] = len(a[0]) + len(a[1])
    print(table)
    return table


def binary_cohens_kappa(a, b, return_agreement_table=True):
    """
    Performs Cohen's kappa statistic on binary data from 2 observers.

    :param a: annotations/labels for observer 1
    :type a: ndarray
    :param b: annotations.labels for observer 2
    :type b: ndarray
    :param return_agreement_table: if True, returns an agreement table along with Kappa
    :type return_agreement_table: bool (default=True)
    :return: kappa value or kappa value and agreement table (if return_agreement_table is True)
    :rtype: List[float, Optional[ndarray]]
    """
    n_obs1 = len(a)
    n_obs2 = len(b)
    if n_obs1 != n_obs2:
        print('total annotations do not match each other!')
        return
    a.shape = (len(a),)
    b.shape = (len(b),)
    a = a.tolist()
    b = b.tolist()
    yes_obs1 = a.count(1)
    no_obs1 = a.count(0)
    yes_obs2 = b.count(1)
    no_obs2 = b.count(0)
    agreement = [i for i, j in zip(a, b) if i == j]
    yy = agreement.count(1)
    nn = agreement.count(0)
    expected_agreement = ((yes_obs1 / n_obs2) * (yes_obs2 / n_obs2)) + ((no_obs1 / n_obs2) * (no_obs2 / n_obs2))
    observed_agreement = (yy + nn) / n_obs2
    k = (observed_agreement - expected_agreement) / (1 - expected_agreement)
    print('Pe = ' + str(expected_agreement))
    print('Po = ' + str(observed_agreement))
    print('Cohens Kappa = ' + str(k))

    if return_agreement_table:
        table = agreement_table([yes_obs1, no_obs1], [yes_obs2, no_obs2], [yy, nn])
        return k, table
    else:
        return k


def binary_tpfptnfn(predictions, label):
    """
    Returns the number of true positives, false positives,
    true negatives,and false negatives for binary data.

    Keyword arguments:
    predictions -- an array or list of binary (0 or 1) predictions
    labels -- an array or list of binary (0 or 1) ground truth labels
    """
    tp = 0
    fp = 0
    tn = 0
    fn = 0
    for i in range(len(predictions)):
        if predictions[i] == 1 and label[i] == 1:
            tp += 1
        elif predictions[i] == 1 and label[i] == 0:
            fp += 1
        elif predictions[i] == 0 and label[i] == 0:
            tn += 1
        elif predictions[i] == 0 and label[i] == 1:
            fn += 1
        else:
            print('error', 'wrong label?')
            return
    print('TP:', tp)
    print('FP:', fp)
    print('TN:', tn)
    print('FN:', fn)
    return tp, fp, tn, fn


def sensitivity_specificity_ppv_npv(tp, fp, tn, fn):
    """
    Returns sensitivity, specificity, positive predictive value,
    and negative predictive value.

    Keyword arguments:
    tp -- # of true positives
    fp -- # of false positives
    tn -- # of true negatives
    fn -- # of false negatives
    """
    sensitivity = tp / (tp + fn)
    specificity = tn / (tn + fp)
    ppv = tp / (tp + fp)
    npv = tn / (tn + fn)
    print('Sensitivity:', sensitivity)
    print('Specificity:', specificity)
    print('PPV:', ppv)
    print('NPV:', npv)
    return sensitivity, specificity, ppv, npv


def ai_binary_statistics(predictions, labels):
    """
    Returns TP, FP, TN, FN, sensitivity, specificity, PPV, and NPV.

    Keyword arguments:
    predictions -- list or array of binary (0 or 1) predictions
    labels -- binary (0 or 1) ground truth
    """
    if len(predictions) == len(labels):
        a, b, c, d = binary_tpfptnfn(predictions, labels)
        e, f, g, h = sensitivity_specificity_ppv_npv(a, b, c, d)
    else:
        print('prediction-label mismatch!')
        return
    return a, b, c, d, e, f, g, h


def convert_probabilities_to_predictions(probs, from_logits=True):
    """
    Converts a 2D array of logits (default) or probabilities to
    binary predictions.

    Keyword arguments:
    probs -- a 2D array of binary class probabilities or logits
    from_logits -- boolean whether probs values are logits (default True)
    """
    if from_logits:
        return np.argmax(softmax(probs), axis=1)
    else:
        return np.argmax(probs, axis=1)


def get_pred_and_binary_statistics(probs, labels):
    """
    Returns binary comparison statistics (TP, FP, TN, FN, sensitivity,
    specificity, PPV, NPV) given logits and ground truth.

    Keyword arguments:
    probs -- a 2D array of binary class logits
    labels -- an array or list of binary ground truth labels
    """
    pred = convert_probabilities_to_predictions(probs)
    a, b, c, d, e, f, g, h = ai_binary_statistics(pred, labels)
    counter = 0
    for i in range(len(pred)):
        if pred[i] == labels[i]:
            counter += 1
    print('accuracy:', counter / len(pred))
    return a, b, c, d, e, f, g, h


def softmax(logits, axis=-1):
    """
    Returns probabilities from logits.

    Keyword arguments:
    logits -- a 2D array of logit values
    axis -- axis to calculate probabilities along (default)
    """
    y = np.exp(logits - logits.max(axis=axis, keepdims=True))
    return y / y.sum(axis=axis, keepdims=True)


# code way to optionally save the figures
#
#
#
def roc_pr_curves(probabilities, labels, roc=True, pr=True,
                  from_logits=True):
    """
    Returns Receiver operating characteristic (ROC) and precision-recall
    (PR) curves.

    Keyword arguments:
    probabilities -- a 2D array of binary class probabilities or logits
    labels -- an array or list of binary ground truth labels
    ROC -- Boolean; whether to return ROC graph (default True)
    PR -- Boolean; whether to return PR graph (default True)
    from_logits -- Boolean; whether probabilities are logits (default True)
    """
    roc_list = []
    auroc_list = []
    pr_list = []
    avgpr_list = []
    for i in probabilities:
        if from_logits:
            probs = softmax(i)
        else:
            probs = i
        if roc:
            # calculates FPR and TPR for a ROC curve
            roc = sklearn.metrics.roc_curve(labels, probs[:, 1])
            roc_list.append(roc)
            # calculates area under the ROC curve
            auc_roc = sklearn.metrics.roc_auc_score(labels, probs[:, 1])
            auroc_list.append(auc_roc)
        if pr:
            # calculates precision and recall for a PR curve
            pr = sklearn.metrics.precision_recall_curve(labels, probs[:, 1])
            pr_list.append(pr)
            # calculates average precision
            avg_pr = sklearn.metrics.average_precision_score(labels, probs[:, 1])
            avgpr_list.append(avg_pr)
    if roc:
        # plot ROC curve and AUROC
        plt.figure()
        lw = 2
        for i in range(len(roc_list)):
            plt.plot(roc_list[i][0], roc_list[i][1], lw=lw,
                     label='AUROC = %0.4f' % auroc_list[i])
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver operating characteristic curve')
        plt.legend(loc="lower right")
        plt.show()
    if pr:
        # plot PR curve and avg precision
        plt.figure()
        lw = 2
        for i in range(len(pr_list)):
            plt.plot(pr_list[i][1], pr_list[i][0], lw=lw,
                     label='Avg_precision = %0.2f' % avgpr_list[i])
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('Recall (TP/(TP+FN))')
        plt.ylabel('Precision (TP/(TP+FP))')
        plt.title('Precision-Recall curve')
        plt.legend(loc="lower right")
        plt.show()


# Need to add code to add save option for figures
#
#
#
def bland_altman(x1, x2, percent_difference=False, save=False):
    """
    Returns Bland-Altman plots along with mean, min/max 95% CI of the
    mean, standard deviation, and min/max 95% CI of the standard deviation.

    Keyword arguments:
    x1 -- an array or list of values
    x2 -- an array or list of values of equal length as x1
    percent_difference -- Boolean; whether you want the percent_difference
    plotted instead of only the difference
    save -- Boolean; whether to save plot as figure
    """
    difference = [i - j for (i, j) in zip(x1, x2)]
    average = [sum([i, j]) / 2 for (i, j) in zip(x1, x2)]
    if percent_difference:
        difference = [i / j for (i, j) in zip(difference, average)]
    avg, var, std = scipy.stats.bayes_mvs(difference, 0.95)
    plt.figure()
    # points
    plt.scatter(average, difference)
    # mean difference
    plt.plot([0, max(average)], [avg[0], avg[0]], color='orange', label='Mean difference')
    # x origin axis
    plt.plot([0, max(average)], [0, 0], lw=1, color='k')
    # upper 95% CI for mean
    plt.plot([0, max(average)], [avg[1][1], avg[1][1]], color='orange', ls='--')
    # lower 95% CI for mean
    plt.plot([0, max(average)], [avg[1][0], avg[1][0]], color='orange', ls='--')
    # upper 1.96*std
    plt.plot([0, max(average)], [avg[0] + 1.96 * std[0], avg[0] + 1.96 * std[0]], color='lightblue',
             label='+/- 1.96*Stdev')
    # upper 95% CI for upper std
    plt.plot([0, max(average)], [avg[0] + 1.96 * std[1][1], avg[0] + 1.96 * std[1][1]], color='lightblue', ls='--')
    # lower 95% CI for upper std
    plt.plot([0, max(average)], [avg[0] + 1.96 * std[1][0], avg[0] + 1.96 * std[1][0]], color='lightblue', ls='--')
    # lower 1.96*std
    plt.plot([0, max(average)], [avg[0] - 1.96 * std[0], avg[0] - 1.96 * std[0]], color='lightblue')
    # upper 95% CI for lower std
    plt.plot([0, max(average)], [avg[0] - 1.96 * std[1][0], avg[0] - 1.96 * std[1][0]], color='lightblue', ls='--')
    # lower 95% CI for lower std
    plt.plot([0, max(average)], [avg[0] - 1.96 * std[1][1], avg[0] - 1.96 * std[1][1]], color='lightblue', ls='--')

    plt.xlim([0.0, max(average)])
    plt.legend(loc="upper right")
    # TODO: need to add code to add save option
    plt.show()
    return avg[0], avg[1][0], avg[1][1], std[0], std[1][0], std[1][1]


def coordinate_separation(total_cpec_coords, affected_cpec_coords, bin_num=50,
                          plot_range=(5, 20)):
    """
    Returns a histogram showing pixel distances within total CPEC coords
    and distances between total CPEC and affected CPEC coords.

    Histogram can be interpreted to determine a cutoff distance that
    identifies coordinates that are likely the same cell.

    Keyword arguments:
    total_cpec_coords -- a 2D array of all CPEC coordinates
    affected_cpec_coords -- a 2D array of affected CPEC coordinates
    bin_num -- # of bins to plot in the histogram (default 50)
    plot_range -- a two number tuple range of the histogram (default (5,20))
    """
    tc = KDTree(total_cpec_coords)
    self_separation = tc.query(total_cpec_coords, k=2)
    affected_separation = tc.query(affected_cpec_coords, k=1)
    plt.hist([self_separation[0][..., 1], affected_separation[0]],
             bins=bin_num, range=plot_range)


def indices_for_deleting_affected_coords_self(affected_self_separation, threshold):
    del_affected = []
    for i in range(len(affected_self_separation[0][..., 1])):
        if affected_self_separation[0][..., 1][i] <= threshold:
            del_affected.append(affected_self_separation[1][..., 1][i])
    return del_affected


def affected_self_total_separation_v1(total_coords, affected_coords):
    """
    Calculates nearest-neighbor (nn) distances.
    :param total_coords: total CPEC coordinates
    :param affected_coords: affected CPEC coordinates
    :return: nn distances within affected coords and between affected and total coords
    """
    ac = KDTree(affected_coords)
    tc = KDTree(total_coords)
    affected_self_separation = ac.query(affected_coords, k=2)
    affected_total_separation = tc.query(affected_coords, k=1)
    return affected_self_separation, affected_total_separation


def remove_redundant_affected_coords(total_cpec_coords, affected_cpec_coords,
                                     threshold):
    """
    Returns a 2D array of affected coordinates with redundant coordinates removed.

    Only redundant coordinates within affected coordinates are removed.

    Keyword arguments:
    total_cpec_coords -- a 2D array of all CPEC coordinates
    affected_cpec_coords -- a 2D array of affected CPEC coordinates
    threshold -- distance cutoff to determine redundant coordinates
    """
    ac = KDTree(affected_cpec_coords)
    affected_self_separation = ac.query(affected_cpec_coords, k=2)
    if total_cpec_coords:
        tc = KDTree(total_cpec_coords)
        affected_total_separation = tc.query(affected_cpec_coords, k=1)

    del_affected = indices_for_deleting_affected_coords_self(affected_self_separation, threshold)
    # May be able to remove this part of the code in the future.
    # That way, total CPEC coordinates do not need to be provided.
    if total_cpec_coords:
        for i in range(len(affected_total_separation[0])):
            if affected_total_separation[0][i] > threshold:
                del_affected.append(i)
    print(len(del_affected))
    trimmed_affected_coords = np.delete(affected_cpec_coords, del_affected, 0)
    return trimmed_affected_coords


def get_unaffected_from_total_coords(total_cpec_coords, trimmed_affected_cpec_coords, threshold):
    """
    Returns an array of unaffected (binary label:0) coordinates from total
    coordinates.

    Keyword arguments:
    total_cpec_coords -- an array of total CPEC coordinates
    trimmed_affected_cpec_coords -- an array of affected CPEC coordinates
    threshold -- distance threshold to determine matching total and affected
    coordinates
    """
    affected_self_separation, affected_total_separation = affected_self_total_separation_v1(total_cpec_coords,
                                                                                            trimmed_affected_cpec_coords
                                                                                            )
    del_total = []
    del_affected = indices_for_deleting_affected_coords_self(affected_self_separation, threshold)
    for i in range(len(affected_total_separation[0])):
        if affected_total_separation[0][i] > threshold:
            del_affected.append(i)
        else:
            del_total.append(affected_total_separation[1][i])
    print(len(del_affected), len(del_total))
    unaffected_coords = np.delete(total_cpec_coords, del_total, 0)
    return unaffected_coords


# CODE not finished; Need to finish
# def remove_redundant_coords(main_list, check_listk, radius):
#    """
#    Returns affected and total CPEC coordinates with redundant coordinates
#    removed.
#
#    Keyword arguments:
#    main_list --
#    check_;ost
#    """
#    main = []
#    check = []
#    for i in range(len(nn_dist_affected[0])):
#        if nn_dist_affected[0][i] > 15:
#            del_affected.append(i)
#        else:
#            del_total.append(nn_dist_affected[1][i])
#    print(len(del_total), len(del_affected))


def cpec_density(total_cpec_coords, radius=310.559):
    """

    :param total_cpec_coords:
    :type total_cpec_coords:
    :param radius:
    :type radius:
    :return:
    :rtype:
    """
    tc = KDTree(total_cpec_coords)
    total = tc.query_radius(total_cpec_coords, r=radius, count_only=True)
    return


def percent_affected(total_cpec_coords, affected_cpec_coords, density=False, radius=310.559):
    """
    Returns a list of percent affected values (within a given radius) for
    use in applying a heatmap to a scatterplot.

    Keyword arguments:
    total_cpec_coords -- a 2D array of all CPEC coordinates
    affected_cpec_coords -- a 2D array of affected CPEC coordinates
    radius -- radius to calculate percent of affected CPEC (default 310.559)
    """
    ac = KDTree(affected_cpec_coords)
    tc = KDTree(total_cpec_coords)

    total = tc.query_radius(total_cpec_coords, r=radius, count_only=True)
    affected = ac.query_radius(total_cpec_coords, r=radius, count_only=True)
    percentage_affected = affected / total
    for i in range(len(percentage_affected)):
        if percentage_affected[i] > 1:
            percentage_affected[i] = 1
    if density:
        return percentage_affected, total
    else:
        return percentage_affected


def plot_xy_hist(total_cpec_coords, percentage_affected, xy_savefile,
                 hist_savefile, xy=True, hist=True, axes=True):
    """
    Returns and saves a xy plot of CPEC coordinates with local percent
    affected heatmap and histogram.

    Keyword arguments:
    total_cpec_coords -- a 2D array of all CPEC coordinates
    percent_affected -- a list of percent affected values within a radius
    xy_savefile -- string; save filename for the xy plot
    hist_savefile -- string; save filename for the histogram
    XY -- Boolean; whether to plot a xy plot (default True)
    HIST -- Boolean; whether to plot a histogram (default True)
    axes -- Boolean; whether to plot axes for the xy plot
    """
    z = percentage_affected
    if xy:
        # Plot XY plot

        # Plot CPEC coords
        x = total_cpec_coords[:, 0]
        y = total_cpec_coords[:, 1]

        # local percent affected heatmap
        fig, ax = plt.subplots()
        ax.scatter(x, y, c=z, cmap='inferno', s=100, marker=".", edgecolor='')
        plt.axis('scaled')
        if not axes:
            plt.axis('off')
        plt.gca().invert_yaxis()
        plt.autoscale(False)
        default_size = fig.get_size_inches()
        fig.set_size_inches((default_size[0] * 5, default_size[1] * 5))
        plt.savefig(xy_savefile, transparent=True, bbox_inches='tight', pad_inches=0)
    if hist:
        # Plot Histogram
        font = {'family': 'normal',
                'weight': 'bold',
                'size': 22}
        matplotlib.rc('font', **font)
        plt.figure(figsize=(15, 15))
        plt.hist(z, bins=23, range=(0, 1), ec='black')
        plt.savefig(hist_savefile, transparent=True, bbox_inches='tight', pad_inches=0)


def montecarlo_histogram_overlay(x, y, filename):
    """
    Generates a overlay histogram.
    :param filename: save filename
    :type filename:
    :param x: observed percent affected values
    :type x:
    :param y: montecarlo percent affected values
    :type y:
    :return:
    :rtype:
    """
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 22}
    matplotlib.rc('font', **font)
    fig, ax1 = plt.subplots(figsize=(15, 15))

    ax1.hist(x, bins=23, range=(0, 1), label='Observed', alpha=0.9, ec='black')
    ax2 = ax1.twinx()
    ax2.hist(y.flatten(), bins=23, range=(0, 1), color='tab:orange', label='Monte Carlo', alpha=0.7, ec='black')
    fig.legend(loc=(0.7, 0.85))

    plt.savefig(filename, transparent=True, bbox_inches='tight', pad_inches=0)


def homogeneity_monte_carlo(affected, total, iterations, radius=310.559):
    """
    Returns monte_carlo simulations of local percent affected values.

    Keyword arguments:
    affected -- a 2D array of affected CPEC coordinates
    total -- a 2D array of all CPEC coordinates
    iterations -- # of simulations to perform
    radius -- radius to calculate percent of affected CPEC (default 310.559)
    """
    simulated_densities = []
    k = len(affected)
    totallist = total.tolist()
    counter = 0
    for i in range(iterations):
        sim_affected = random.sample(totallist, k)
        sim_affected = np.stack(sim_affected)
        affected_tree = KDTree(sim_affected)
        total_tree = KDTree(total)
        nearby_affected = affected_tree.query_radius(total, r=radius, count_only=True)
        nearby_total = total_tree.query_radius(total, r=radius, count_only=True)
        density = nearby_affected / nearby_total
        simulated_densities.append(density)
        counter += 1
        print(counter)
    simulated_densities = np.stack(simulated_densities, axis=0)
    return simulated_densities


def sort_affected_coords_from_aipredictions(aipredictions, coords):
    """
    Sorts out affected coordinates from total coordinates based on
    ai predictions

    Keyword arguments:
    aipredictions -- a list or array of binary predictions (1=affected)
    coords -- an array of coordinates
    """
    affected = []
    for i in range(len(aipredictions)):
        if aipredictions[i] == 1:
            affected.append(i)
    affected_coords = coords[affected]
    return affected_coords


def kolmogorov_smirnov_statistic_for_wsi_montecarlo_simulations(obs_pa, simulated_pa, start=None, stop=None,
                                                                observed_only=False, test_observed=False):
    sim_dvalues = []
    sim_pvalues = []
    observed = np.load(obs_pa)
    obs_reshape = observed.copy()
    obs_reshape.shape = (1, observed.shape[0])
    simulations = np.load(simulated_pa)
    obs_and_sims = np.concatenate([obs_reshape, simulations]).flatten()
    counter = 0
    if test_observed:
        obs_dvalue, obs_pvalue = scipy.stats.ks_2samp(observed, obs_and_sims)
        print('Observed data D-value:', obs_dvalue, 'Observed data P-value:', obs_pvalue)
        if observed_only:
            return obs_dvalue, obs_pvalue
    if not start:
        start = 0
    if not stop:
        stop = len(simulations)
    for i in range(start, stop):
        d, p = scipy.stats.ks_2samp(simulations[i], obs_and_sims)
        sim_dvalues.append(d)
        sim_pvalues.append(p)
        counter += 1
        print(counter, 'D-value:', sim_dvalues[counter - 1], 'P-value:', sim_pvalues[counter - 1])
    if test_observed:
        return obs_dvalue, obs_pvalue, sim_dvalues, sim_pvalues
    else:
        return sim_dvalues, sim_pvalues


def plot_training(history):
    """
    Diplays a training history graph based on accuracy. Good for classifiers.

    Keyword arguments:
    history: history object that keras returns during training.
    """
    # Plot training & validation accuracy values
    plt.plot(history.history['accuracy'])
    plt.plot(history.history['val_accuracy'])

    plt.title('Model accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Test'], loc='upper left')
    plt.show()

    # Plot training & validation loss values
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('Model loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Test'], loc='upper left')
    plt.show()


def retinanet_ious(validation_data_dict, model, boundingbox):
    ious = {
        'med': [],
        'p25': [],
        'p75': []}

    box = model.predict(validation_data_dict)
    if type(box) is list:
        box = {k: l for k, l in zip(model.output_names, box)}

    pred_anchors, _ = boundingbox.convert_box_to_anc(box)
    true_anchors, _ = boundingbox.convert_box_to_anc(validation_data_dict)

    curr = []
    for pred, true in zip(pred_anchors, true_anchors):
        for p in pred:
            iou = boundingbox.calculate_ious(box=p, anchors=true)
            if iou.size > 0:
                curr.append(np.max(iou))
            else:
                curr.append(0)
    if len(curr) == 0:
        curr = [0]

    ious['med'].append(np.median(curr))
    ious['p25'].append(np.percentile(curr, 25))
    ious['p75'].append(np.percentile(curr, 75))

    return {k: np.array(v) for k, v in ious.items()}


def pca_analysis(data, n_comp=3, graph=True, pd_database=False):
    # TODO: add capability to provide a numpy array for data
    # assumes 0 index column is sample_name/Target
    if pd_database:
        df = data
    else:
        df = pd.read_excel(data)
    morphologies = list(df.columns)[1:]
    x = df.loc[:, morphologies].values
    y = df.loc[:, [df.columns[0]]].values
    pca = sklearn.decomposition.PCA(n_components=n_comp)
    principalcomponents = pca.fit_transform(x)
    component_names = [f'principal component {i+1}' for i in range(n_comp)]
    principal_df = pd.DataFrame(data=principalcomponents, columns=component_names)
    final_df = pd.concat([principal_df, df[[df.columns[0]]]], axis=1)
    if graph:
        targets = list(np.unique(df[df.columns[0]]))
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1)
        ax.set_xlabel('principal component 1', fontsize=15)
        ax.set_ylabel('principal component 2', fontsize=15)
        ax.set_title('2 component PCA', fontsize=20)
        for target in targets:
            indices_to_keep = final_df[df.columns[0]] == target
            ax.scatter(final_df.loc[indices_to_keep, component_names[0]],
                       final_df.loc[indices_to_keep, component_names[1]],
                       s=50)
        ax.legend(targets, bbox_to_anchor=(1, 1), loc='upper left', ncol=1)
        ax.grid()
        return final_df, fig
    else:
        return final_df


def plot_pca(data, pc, _3d=False, elev=None, azim=None, labels=False, legend=True):
    # data must be a pandas dataframe with principal components and targets
    # pc is a tuple of which principal components to plot (1 index)
    targets = list(np.unique(data['Target']))
    fig = plt.figure(figsize=(8, 8))
    if _3d:
        ax = fig.add_subplot(1, 1, 1, projection='3d')
        ax.set_xlabel(f'principal component {pc[0]}', fontsize=15)
        ax.set_ylabel(f'principal component {pc[1]}', fontsize=15)
        ax.set_zlabel(f'principal component {pc[2]}', fontsize=15)
        ax.set_title('3 component PCA', fontsize=20)
        for target in targets:
            indices_to_keep = data['Target'] == target
            ax.scatter(data.loc[indices_to_keep, f'principal component {pc[0]}'],
                       data.loc[indices_to_keep, f'principal component {pc[1]}'],
                       data.loc[indices_to_keep, f'principal component {pc[2]}'],
                       s=50)
            if labels:
                label = str(data.loc[indices_to_keep, ['Target']].values[0][0])
                ax.annotate(label, (data.loc[indices_to_keep, f'principal component {pc[0]}'],
                                    data.loc[indices_to_keep, f'principal component {pc[1]}']),
                            textcoords='offset points',
                            xytext=(0, 10),
                            ha='center')
        if legend:
            ax.legend(targets, bbox_to_anchor=(1, 1), loc='upper left', ncol=1)
        ax.grid()
        ax.view_init(elev=elev, azim=azim)
    else:
        ax = fig.add_subplot(1, 1, 1)
        ax.set_xlabel(f'principal component {pc[0]}', fontsize=15)
        ax.set_ylabel(f'principal component {pc[1]}', fontsize=15)
        ax.set_title('2 component PCA', fontsize=20)
        for target in targets:
            indices_to_keep = data['Target'] == target
            ax.scatter(data.loc[indices_to_keep, f'principal component {pc[0]}'],
                       data.loc[indices_to_keep, f'principal component {pc[1]}'],
                       s=50)
            if labels:
                label = str(data.loc[indices_to_keep, ['Target']].values[0][0])
                ax.annotate(label, (data.loc[indices_to_keep, f'principal component {pc[0]}'],
                                    data.loc[indices_to_keep, f'principal component {pc[1]}']),
                            textcoords='offset points',
                            xytext=(0, 10),
                            ha='center')
        if legend:
            ax.legend(targets, bbox_to_anchor=(1, 1), loc='upper left', ncol=1)
        ax.grid()
    return fig



import os
import glob
import re
import numpy as np
from skimage import io
import skimage
import skimage.morphology
import skimage.filters
import skimage.transform
import scipy
from scipy import ndimage
import openslide
import matplotlib.pyplot as plt
import pylab
import cv2 as cv
from tensorflow import keras
from sklearn.neighbors import KDTree
import h5py
import pickle


def load_tif(tif_file):
    """
    Returns array of a .tif image.

    Keyword arguments:
    tif_file -- string; .tif filename
    """
    return io.imread(tif_file)


# Test to see if i can just directly return the ndimage.label function
#
def label_mask(mask):
    """
    Returns an array where each mask object has a unique label.

    Keyword arguments:
    mask -- an array of a binary mask
    """
    label, n = ndimage.label(mask)
    return label, n


def get_slices(label):
    """
    Returns slices to isolate and extract a labeled mask object.

    Keyword arguments:
    label -- a labeled mask array (usually output from label_mask())
    """
    x = ndimage.find_objects(label)
    return x


def get_padded_slices_v2(slices, pad, label):
    """
    Adds a buffer to object slices.

    Keyword arguments:
    slices -- list of slices for a single object (usually from get_slices())
    pad -- amount in pixel to pad/buffer the slices
    label -- a labeled mask array (usually output from label_mask())
    """
    new_slices = []
    for i, s in enumerate(slices):
        start = s.start
        stop = s.stop
        if i == 3:
            # code to prevent slicing of the channels
            new_start = 0
            new_stop = 3
            new_slices.append(slice(new_start, new_stop))
            continue

        if start - pad < 0:
            new_start = 0
        else:
            new_start = start - pad

        if stop + pad > label.shape[i]:
            new_stop = label.shape[i]
        else:
            new_stop = stop + pad

        if new_stop - new_start > 64:
            break
        else:
            new_slices.append(slice(new_start, new_stop))
    return tuple(new_slices)


def pad_to_64(cropped_shape, cropped_image):
    """
    Takes a cropped image and pads it to 64 pixel in each dimension.

    Keyword arguments:
    cropped_shape -- tuple; array shape of cropped_image
    cropped_image -- an array of an image
    """
    a, b, c, d, e, f = 0, 0, 0, 0, 0, 0
    for i in range(0, 3):
        if cropped_shape[i] % 2 == 1:
            if i == 0:
                a = int((64 - cropped_shape[i] + 1) / 2)
                b = a - 1

            elif i == 1:
                c = int((64 - cropped_shape[i] + 1) / 2)
                d = c - 1

            elif i == 2:
                e = int((64 - cropped_shape[i] + 1) / 2)
                f = e - 1

        else:
            if i == 0:
                a = int((64 - cropped_shape[i]) / 2)
                b = int((64 - cropped_shape[i]) / 2)

            elif i == 1:
                c = int((64 - cropped_shape[i]) / 2)
                d = int((64 - cropped_shape[i]) / 2)

            elif i == 2:
                e = int((64 - cropped_shape[i]) / 2)
                f = int((64 - cropped_shape[i]) / 2)
    padded_cropped_image = np.pad(cropped_image, ((a, b), (c, d), (e, f), (0, 0)), 'constant')
    return padded_cropped_image


def extract_padded_cropped_inclusions(mask_file, image_file, save_file, pad=5):
    """
    Extracts images of individual objects based on a binary mask.

    :param mask_file: filename of mask image
    :type mask_file: str
    :param image_file: filename of image
    :type image_file: str
    :param save_file: save filename
    :type save_file: str
    :param pad: amount of pixel padding when slicing out objects based on mask
    :type pad: int
    :return: an array of 3D padded cropped inclusions
    :rtype: ndarray
    """
    mask = load_tif(mask_file)
    original_image = load_tif(image_file)
    label, n = label_mask(mask)
    slices = get_slices(label)
    counter = 0
    images = []
    for i in range(n):
        new_slices = get_padded_slices_v2(slices[i], pad, label)
        if len(new_slices) < 4:
            continue
        else:
            cropped_image = original_image[new_slices]
            cropped_shape = cropped_image.shape
            images.append(pad_to_64(cropped_shape, cropped_image))
            counter += 1
    images = np.stack(images, axis=0)
    print('Number of extracted images:', len(images))
    np.save(save_file, images)
    return images


def randomize_and_segregate_dataset(images, labels):
    """
    Shuffles and splits images and their labels into 3 datasets:
    training, validation, and test datasets

    Keyword arguments:
    images -- an array of images
    labels -- an array of labels
    """
    rand_indices = np.random.choice(len(images), size=(np.rint((len(images) * 0.2)).astype('int64')), replace=False)
    test_batch = images[rand_indices]
    test_labels = labels[rand_indices]
    training_and_validation = np.delete(images, rand_indices, axis=0)
    training_and_validation_labels = np.delete(labels, rand_indices, axis=0)
    rand_indices2 = np.random.choice(len(training_and_validation),
                                     size=(np.rint((len(training_and_validation) * 0.2)).astype('int64')),
                                     replace=False)
    validation_batch = training_and_validation[rand_indices2]
    validation_labels = training_and_validation_labels[rand_indices2]
    training = np.delete(training_and_validation, rand_indices2, 0)
    traininglabels = np.delete(training_and_validation_labels, rand_indices2, 0)
    # below is new code to shuffle the training data
    rand_indices3 = np.random.choice(len(training), size=(len(training)), replace=False)
    training_batch = training[rand_indices3]
    training_labels = traininglabels[rand_indices3]
    return training_batch, training_labels, validation_batch, validation_labels, test_batch, test_labels


def load_wsi_coordinates(list_of_text_files):
    """
    Loads a list of WSI mapping coordinates data and returns a single array.

    Keyword arguments:
    list_of_text_files -- a list containing one or more .txt filesnames
    """
    files = []
    for i in list_of_text_files:
        x = (np.loadtxt(i)).astype(int)
        if len(x.shape) == 2:
            x = x[:, 1:3]
        elif len(x.shape) == 1:
            x.shape = (1, 3)
            x = x[:, 1:3]
        else:
            print('error')
            return
        files.append(x)
    combined_list = np.concatenate(files, axis=0)
    return combined_list


def load_wsi_coords_and_labels(coords_list, defined_label):
    """
    Loads a list of WSI mapping coordinates data and returns a single
    array of coordinates and an array of labels

    Keyword arguments:
    list_of_text_files -- a list containing one or more .txt filesnames
    defined_label -- a list of 3 binary label values ([0,0,0])
    """
    x = load_wsi_coordinates(coords_list)
    labels = np.zeros((len(x), 3), dtype=int)
    labels[:] = defined_label
    return x, labels


def wsi_cell_extraction_from_coords_v3(wsi_image_filename, im_size, coords):
    """
    Extracts cropped images at given coordinates from a larger image.

    Keyword arguments:
    wsi_image_filename -- string: filename of the image to extract from
    im_size -- size in pixels for extracted images
    coords -- list or array of coordinates in order
    """
    wsi = openslide.open_slide(wsi_image_filename)
    cells = []
    counter = 0
    for i in coords:
        top_left_row = int(i[1] - (im_size / 2))
        top_left_column = int(i[0] - (im_size / 2))
        cells.append(wsi.read_region((top_left_column, top_left_row), 0, (im_size, im_size)))
        counter += 1
        print(counter, "out of", len(coords), "slices")
    cells = np.stack(cells, axis=0)
    cells = cells[:, :, :, :-1]
    return cells


def replace_w_b(x):
    """
    Replaces 'white' pixels with black.

    Keyword arguments:
    x -- an image array or array of images
    """
    counter = 0
    if len(x.shape) == 3:
        print('Single input image')
        for j in range(x.shape[0]):
            for k in range(x.shape[1]):
                if (x[j, k, :] == np.array([255, 255, 255])).all():
                    x[j, k, :] = 0
    elif len(x.shape) == 4:
        print('Multiple input images')
        for i in range(len(x)):
            for j in range(x[i].shape[0]):
                for k in range(x[i].shape[1]):
                    if (x[i, j, k, :] == np.array([255, 255, 255])).all():
                        x[i, j, k, :] = 0
                        counter += 1
                        print(counter, i, 'out of', len(x))
    return x


def extract_im_and_labels_from_image(list_of_annotation_files, image_filename, im_size=64, from_mosaic=True):
    """
    Extracts cropped images and labels from a larger image (image mosaic
    or WSI) and annotation files.

    Keyword arguments:
    list_of_annotation_files -- list of strings; annotation filenames (.txt)
    image_filename -- string; filename of image mosaic or whole slide image
    im_size -- size in pixels of images to extract
    from_mosaic -- Boolean; whether source image is a mosaic or WSI; for
    determining whether to replace 'white' border pixels with black
    """
    counter = 0
    for i in list_of_annotation_files:
        if '_un.txt' in i:
            coords_u, labels_u = load_wsi_coords_and_labels([i], [0, 0, 0])
            counter += 1
        elif '_bi.txt' in i:
            coords_b, labels_b = load_wsi_coords_and_labels([i], [1, 0, 0])
            counter += 1
        elif '_v++.txt' in i:
            if '_bi_v++.txt' in i:
                coords_bvv, labels_bvv = load_wsi_coords_and_labels([i], [1, 0, 0])
                counter += 1
            else:
                coords_vv, labels_vv = load_wsi_coords_and_labels([i], [1, 0, 0])
                counter += 1
        elif '_mu.txt' in i:
            if '_bi_mu.txt' in i:
                coords_bm, labels_bm = load_wsi_coords_and_labels([i], [1, 1, 0])
                counter += 1
            else:
                coords_m, labels_m = load_wsi_coords_and_labels([i], [0, 1, 0])
                counter += 1
        elif '_v+.txt' in i:
            if '_bi_v+.txt' in i:
                coords_bv, labels_bv = load_wsi_coords_and_labels([i], [1, 0, 1])
                counter += 1
            elif '_mu_v+.txt' in i:
                if '_bi_mu_v+.txt' in i:
                    coords_bmv, labels_bmv = load_wsi_coords_and_labels([i], [1, 1, 1])
                    counter += 1
                else:
                    coords_mv, labels_mv = load_wsi_coords_and_labels([i], [0, 1, 1])
                    counter += 1
            else:
                coords_v, labels_v = load_wsi_coords_and_labels([i], [0, 0, 1])
                counter += 1
        elif '_vr.txt' in i:
            if '_bi_vr.txt' in i:
                coords_bvr, labels_bvr = load_wsi_coords_and_labels([i], [1, 0, 1])
                counter += 1
            elif '_mu_vr.txt' in i:
                if '_bi_mu_vr.txt' in i:
                    coords_bmvr, labels_bmvr = load_wsi_coords_and_labels([i], [1, 1, 1])
                    counter += 1
                else:
                    coords_mvr, labels_mvr = load_wsi_coords_and_labels([i], [0, 1, 1])
                    counter += 1
            else:
                coords_vr, labels_vr = load_wsi_coords_and_labels([i], [0, 0, 1])
                counter += 1
        else:
            print(i, 'is not properly named/recognized')

    if 'coords_bvv' or 'coords_vv' in locals():
        bvvc = []
        bvvl = []
        if 'coords_bvv' in locals():
            bvvc.append(coords_bvv)
            bvvl.append(labels_bvv)
        if 'coords_vv' in locals():
            bvvc.append(coords_vv)
            bvvl.append(labels_vv)
        if 'coords_b' in locals():
            bvvc.append(coords_b)
            bvvl.append(labels_b)
        if len(bvvc) != 0:
            coords_b = np.concatenate(bvvc, axis=0)
            labels_b = np.concatenate(bvvl, axis=0)
    if 'coords_vr' in locals():
        vc = []
        vl = []
        if 'coords_vr' in locals():
            vc.append(coords_vr)
            vl.append(labels_vr)
        if 'coords_v' in locals():
            vc.append(coords_v)
            vl.append(labels_v)
        if len(vc) != 0:
            coords_v = np.concatenate(vc, axis=0)
            labels_v = np.concatenate(vl, axis=0)
    if 'coords_bvr' in locals():
        bvc = []
        bvl = []
        if 'coords_bvr' in locals():
            bvc.append(coords_bvr)
            bvl.append(labels_bvr)
        if 'coords_bv' in locals():
            bvc.append(coords_bv)
            bvl.append(labels_bv)
        if len(bvc) != 0:
            coords_bv = np.concatenate(bvc, axis=0)
            labels_bv = np.concatenate(bvl, axis=0)
    if 'coords_mvr' in locals():
        mvc = []
        mvl = []
        if 'coords_mvr' in locals():
            mvc.append(coords_mvr)
            mvl.append(labels_mvr)
        if 'coords_mv' in locals():
            mvc.append(coords_mv)
            mvl.append(labels_mv)
        if len(mvc) != 0:
            coords_mv = np.concatenate(mvc, axis=0)
            labels_mv = np.concatenate(mvl, axis=0)
    if 'coords_bmvr' in locals():
        bmvc = []
        bmvl = []
        if 'coords_bmvr' in locals():
            bmvc.append(coords_bmvr)
            bmvl.append(labels_bmvr)
        if 'coords_bmv' in locals():
            bmvc.append(coords_bmv)
            bmvl.append(labels_bmv)
        if len(bmvc) != 0:
            coords_bmv = np.concatenate(bmvc, axis=0)
            labels_bmv = np.concatenate(bmvl, axis=0)

    if 'coords_u' not in locals():
        images_u = None
        labels_u = None
    else:
        images_u = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_u)
        if from_mosaic:
            images_u = replace_w_b(images_u)
    if 'coords_b' not in locals():
        images_b = None
        labels_b = None
    else:
        images_b = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_b)
        if from_mosaic:
            images_b = replace_w_b(images_b)
    if 'coords_m' not in locals():
        images_m = None
        labels_m = None
    else:
        images_m = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_m)
        if from_mosaic:
            images_m = replace_w_b(images_m)
    if 'coords_v' not in locals():
        images_v = None
        labels_v = None
    else:
        images_v = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_v)
        if from_mosaic:
            images_v = replace_w_b(images_v)
    if 'coords_bm' not in locals():
        images_bm = None
        labels_bm = None
    else:
        images_bm = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_bm)
        if from_mosaic:
            images_bm = replace_w_b(images_bm)
    if 'coords_bv' not in locals():
        images_bv = None
        labels_bv = None
    else:
        images_bv = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_bv)
        if from_mosaic:
            images_bv = replace_w_b(images_bv)
    if 'coords_mv' not in locals():
        images_mv = None
        labels_mv = None
    else:
        images_mv = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_mv)
        if from_mosaic:
            images_mv = replace_w_b(images_mv)
    if 'coords_bmv' not in locals():
        images_bmv = None
        labels_bmv = None
    else:
        images_bmv = wsi_cell_extraction_from_coords_v3(image_filename, im_size, coords_bmv)
        if from_mosaic:
            images_bmv = replace_w_b(images_bmv)
    print('compare', len(list_of_annotation_files), counter)

    return [images_u, images_b, images_m, images_v, images_bm, images_bv, images_mv, images_bmv], [labels_u, labels_b,
                                                                                                   labels_m, labels_v,
                                                                                                   labels_bm, labels_bv,
                                                                                                   labels_mv,
                                                                                                   labels_bmv]


def merge_images_labels_without_none(im_var_list, label_var_list):
    """
    merges lists of images and labels together.

    Keyword arguments:
    im_var_list -- list of image arrays
    label_var_list -- list of label arrays
    """
    im = [i for i in im_var_list if i is not None]
    label = [i for i in label_var_list if i is not None]
    if len(im) != len(label):
        print("images and labels don't match")
        return
    # print(im)
    if len(im) == 0:
        print('No images to concatenate!')
        return
    images = np.concatenate(im, axis=0)
    labels = np.concatenate(label, axis=0)
    return images, labels


def images_and_full_annotations(list_of_annotations, source_image, im_size=64, from_mosaic=True):
    """
    Returns an array of images and an array of full morphology labels
    for a single image.

    Keyword arguments:
    list_of_annotations -- list of strings; annotation filenames (.txt)
    source_image -- string; filename of image mosaic or whole slide image
    im_size -- size in pixels of images to extract
    from_mosaic -- Boolean; whether source image is a mosaic or WSI; for
    determining whether to replace 'white' border pixels with black
    """
    ims, lbs = extract_im_and_labels_from_image(list_of_annotations, source_image, im_size=im_size,
                                                from_mosaic=from_mosaic)
    all_im, all_lb = merge_images_labels_without_none(ims, lbs)
    print('image dataset shape:', all_im.shape, 'label dataset shape:', all_lb.shape)
    return all_im, all_lb


def binary_annotation_from_full(full_annotations):
    """
    Returns an array of binary labels based on full annotations.

    Keyword arguments:
    full_annotations -- an array of full annotations
    """
    binary = []
    for i in range(len(full_annotations)):
        if full_annotations[i] != 0:
            binary.append(1)
        else:
            binary.append(0)
    binary = np.stack(binary, axis=0)
    return binary


def wsi_tile_extraction(wsi_filename, im_size, sample_name):
    """
    Converts a WSI in to an array of tiles and saves it as a numpy file.

    Keyword arguments:
    WSI_filename -- string; WSI filename
    im_size -- extracted tile size (in pixels)
    sample_name -- string; case number (ex. UCI-12-18 or A17-54)
    """
    wsi = openslide.open_slide(wsi_filename)
    dim = wsi.dimensions
    tiles = []
    # dim[0] = width, dim[1] = height
    im_num = (dim[1] // im_size) * (dim[0] // im_size)
    for i in range(dim[1] // im_size):
        for j in range(dim[0] // im_size):
            # j represent position on x-axis (different from usual which is row #)
            # i represent position on y-axis (different from usual which is column #)
            a = wsi.read_region((j * im_size, i * im_size), 0, (im_size, im_size))
            a = np.array(a)
            tiles.append(a[:, :, :-1])
            print(len(tiles), 'out of', im_num)
    tile_stack = np.stack(tiles, axis=0)
    print(tile_stack.shape)
    filename = sample_name + '_' + str(dim[1] // im_size) + 'rows_' + str(dim[0] // im_size) + 'columns.npy'
    np.save(filename, tile_stack)
    print('File saved as: ' + str(filename))


def wsi_tile_extraction_v2(wsi_filename, im_size, sample_name, half_res=False):
    """
    Converts a WSI in to an array of tiles and saves it as a numpy file.

    Keyword arguments:
    WSI_filename -- string; WSI filename
    im_size -- extracted tile size (in pixels) before halving resolution(if TRUE)
    sample_name -- string; case number (ex. UCI-12-18 or A17-54)
    half_res -- boolean; if true, halves the image resolution and size
    """
    wsi = openslide.open_slide(wsi_filename)
    dim = wsi.dimensions
    tiles = []
    # dim[0] = width, dim[1] = height
    im_num = (dim[1] // im_size) * (dim[0] // im_size)
    for i in range(dim[1] // im_size):
        for j in range(dim[0] // im_size):
            # j represent position on x-axis (different from usual which is row #)
            # i represent position on y-axis (different from usual which is column #)
            a = wsi.read_region((j * im_size, i * im_size), 0, (im_size, im_size))
            a = np.array(a)
            tiles.append(a[:, :, :-1])
            print(len(tiles), 'out of', im_num)
    tile_stack = np.stack(tiles, axis=0)
    if half_res:
        model = half_tile_resolution(im_size)
        tile_stack = model.predict(tile_stack)
        half_res_string = '_half_resolution'
    else:
        half_res_string = ''
    print(tile_stack.shape)
    if tile_stack.shape[1] != 1024:
        imsize = '_' + str(tile_stack.shape[1]) + 'imsize' + half_res_string
    else:
        imsize = ''
    filename = sample_name + imsize + '_' + str(dim[1] // im_size) + 'rows_' + str(dim[0] // im_size) + 'columns.npy'
    np.save(filename, tile_stack)
    print('File saved as: ' + str(filename))


def half_tile_resolution(im_size):
    inputs = keras.Input(shape=(im_size, im_size, 3))
    out = keras.layers.AveragePooling2D()(inputs)
    model = keras.Model(inputs=inputs, outputs=out)
    return model


def generate_mosaic_template(filename):
    x = np.ones((10249, 10249, 3), dtype=np.uint8)
    x[:, :, :] = 255
    tiles = np.load(filename)
    rand_sample = np.random.choice(len(tiles), 500, replace=False)
    return x, tiles, rand_sample


def wsi_mosaic_start_stop_slices(counter):
    r = counter // 10
    c = counter % 10
    r_start = 1025 * r
    r_stop = 1025 * r + 1024
    c_start = 1025 * c
    c_stop = 1025 * c + 1024
    return r_start, r_stop, c_start, c_stop


def produce_mosaic_separated_rgb(filename, rgb=True):
    """
    Generates a random 10x10 mosaic from extracted WSI tiles along with
    metadata.

    Keyword arguments:
    filename -- string: .npy filename of 1024x1024 pixel WSI tiles
    rgb -- Boolean; whether to generate red, green, blue monochannel
    images in addition to the full rgb image
    """
    x, tiles, rand_sample = generate_mosaic_template(filename)
    y = []
    counter = 0
    for i in rand_sample:
        if np.amax(tiles[i, :, :, 2]) >= 85:
            r_start, r_stop, c_start, c_stop = wsi_mosaic_start_stop_slices(counter)
            x[r_start:r_stop, c_start:c_stop, :] = tiles[i]
            y.append(i)
            counter = counter + 1
            if counter >= 100:
                break
        else:
            continue
    plt.imshow(x[:, :, :])
    skimage.io.imsave(filename[:-21] + '_10x10_sample_mosaic.tif', x)
    np.save(filename[:-21] + '_10x10_sample_mosaic_metadata.npy', y)
    if rgb:
        skimage.io.imsave(filename[:-21] + '_10x10_sample_mosaic_red.tif', x[..., 0])
        skimage.io.imsave(filename[:-21] + '_10x10_sample_mosaic_green.tif', x[..., 1])
        skimage.io.imsave(filename[:-21] + '_10x10_sample_mosaic_blue.tif', x[..., 2])


def produce_additional_mosaic_separated_rgb(filename, rgb=True):
    """
    Generates an additional random 10x10 mosaic from extracted WSI tiles
    along with metadata.

    Keyword arguments:
    filename -- string: .npy filename of 1024x1024 pixel WSI tiles
    rgb -- Boolean; whether to generate red, green, blue monochannel
    images in addition to the full rgb image

    """
    case = re.search('(.+?)_(.+?)rows_(.+?)columns.npy', filename).group(1)
    previous_metadata = glob.glob(case + '*metadata*')
    prev_im = []
    for i in previous_metadata:
        prev_im.extend(np.load(i))

    x, tiles, rand_sample = generate_mosaic_template(filename)
    y = []
    counter = 0
    for i in rand_sample:
        if np.amax(tiles[i, :, :, 2]) >= 85 and i not in prev_im:
            r = counter // 10
            c = counter % 10
            r_start = 1025 * r
            r_stop = 1025 * r + 1024
            c_start = 1025 * c
            c_stop = 1025 * c + 1024
            x[r_start:r_stop, c_start:c_stop, :] = tiles[i]
            y.append(i)
            counter = counter + 1
            if counter >= 100:
                break
        elif np.amax(tiles[i, :, :, 2]) >= 85 and i in prev_im:
            print('found match...skipping', i)
        else:
            continue
    plt.imshow(x[:, :, :])
    skimage.io.imsave(case + '_10x10_sample_mosaic_' + str(len(previous_metadata) + 1) + '.tif', x)
    np.save(case + '_10x10_sample_mosaic_metadata_' + str(len(previous_metadata) + 1) + '.npy', y)
    if rgb:
        skimage.io.imsave(case + '_10x10_sample_mosaic_' + str(len(previous_metadata) + 1) + '_red.tif', x[..., 0])
        skimage.io.imsave(case + '_10x10_sample_mosaic_' + str(len(previous_metadata) + 1) + '_green.tif', x[..., 1])
        skimage.io.imsave(case + '_10x10_sample_mosaic_' + str(len(previous_metadata) + 1) + '_blue.tif', x[..., 2])


def data_augmentation(inputs, labels, iterations, dimensions, rotation=True, flip=True):
    """
    Generates augmented images by randomly flipping and rotating.

    Works for both 2D and 3D images

    Keyword arguments:
    inputs -- an array of images to augment
    labels -- an array of corresponding image labels
    iterations -- how many augmented images to generate per image
    dimensions -- 2 or 3; The number of image dimensions (do not
    include channels dimension)
    rotation -- Boolean; whether to randomly rotate images along each axis
    flip -- Boolean; whether to randomly flip images
    """
    augmented_data = []
    augmented_data_labels = []
    y = [(0, 1), (0, 2), (1, 2)]
    counter = 0
    for i in range(len(inputs)):
        for j in range(iterations):
            if flip:
                x = np.random.choice([0, 1, (0, 1), None])
                if x is not None:
                    im = np.flip(inputs[i], axis=x)
                else:
                    im = inputs[i]
            else:
                im = inputs[i]
            if rotation:
                if dimensions in (2, 3):
                    if dimensions == 2:
                        dimensions = 1
                    for k in range(dimensions):
                        im = scipy.ndimage.rotate(im, np.random.randint(360), axes=y[k], reshape=False)
                    else:
                        print('Provided number of dimensions is invalid! Should be either 2 or 3.')
                        return
            if not flip and rotation:
                print('What do you want me to do if no flipping or rotating?')
            augmented_data.append(im)
            augmented_data_labels.append(labels[i])
        counter += 1
        print('finished', counter, 'out of', len(inputs))
    augmented_data = np.stack(augmented_data, axis=0)
    augmented_data_labels = np.stack(augmented_data_labels, axis=0)
    return augmented_data, augmented_data_labels


def annotation_tool(name, filename, x):
    # May remove the follow if statement. Maybe not worth trying to recover a interrupted run (y).
    #
    #
    # Not sure if i should be checking in both locals and globals.
    if 'y' in locals() or 'y' in globals():
        answer1 = None
        while answer1 not in ("y", "n"):
            answer1 = input("Would you like rerun the annotation tool? This will delete unsaved annotaions! y(yes) or "
                            "n(no)?")
            if answer1 == "y":
                y = []
                j = 0
                break
            elif answer1 == "n":
                return
            else:
                print("Please enter y or n")
    # Check for existing in progress annotation file###
    save_list = glob.glob(name + '_annotations_*_out_of_*.npy')
    save_list_complete = glob.glob(name + '_annotations.npy')
    if save_list_complete:
        answer3 = None
        while answer3 not in ("y", "n"):
            answer3 = input("There is an existing completed annotation file! Would you like to continue? y(yes) or n("
                            "no)?")
            if answer3 == "y":
                y = []
                j = 0
            elif answer3 == "n":
                return
            else:
                print("Please enter y or n")
    elif save_list:
        answer2 = None
        while answer2 not in ("y", "n"):
            answer2 = input("Would you like to continue from save file! y(yes) or n(no)?")
            if answer2 == "y":
                y = (np.load(save_list[0]))
                y = y.tolist()
                j = len(y)
            elif answer2 == "n":
                y = []
                j = 0
            else:
                print("Please enter y or n")
    else:
        y = []
        j = 0
    while j < len(x):
        fig = pylab.figure()
        default_size = fig.get_size_inches()
        fig.set_size_inches((default_size[0] * 3, default_size[1] * 3))
        for i in range(3):
            if i == 0:
                k = 0
            elif i == 1:
                k = 1
            else:
                k = 2
            z = x[j]
            z = z[:, :, :, ::-1]
            im = skimage.transform.resize(skimage.img_as_ubyte(skimage.exposure.equalize_adapthist(
                skimage.exposure.rescale_intensity(np.max(z, axis=k), in_range='uint12'))), (512, 512, 3), order=3)
            im[:, :, 2] = 0
            fig.add_subplot(1, 3, i + 1)
            pylab.imshow(im)
            pylab.axis('off')
        print('image #' + str(j + 1) + ' out of ' + str(len(x)))
        pylab.show()
        answer = None
        while answer not in ("y", "n", "u", "s", "ss"):
            answer = input("Is this an inclusion? y(yes), n(no), u(undo prev), s(save), ss(save and stop)")
            if answer == "y":
                y.append(1)
                j = j + 1
            elif answer == "n":
                y.append(0)
                j = j + 1
            elif answer == "u":
                del y[-1]
                j = j - 1
            elif answer == "s":
                save_list = glob.glob(name + '_annotations_*_out_of_*.npy')
                if j == 0:
                    print('Nothing to save!')
                elif save_list:
                    os.remove(save_list[0])
                d = np.asarray(y)
                np.save(name + '_annotations_' + str(j) + '_out_of_' + str(len(x)) + '.npy', d)
            elif answer == "ss":
                save_list = glob.glob(name + '_annotations_*_out_of_*.npy')
                if j == 0:
                    print('Nothing to save! Quiting...')
                    return
                elif save_list:
                    os.remove(save_list[0])
                d = np.asarray(y)
                np.save(name + '_annotations_' + str(j) + '_out_of_' + str(len(x)) + '.npy', d)
                return
            else:
                print("Please enter y, n, u, s, or ss.")
    p = np.asarray(y)
    d = np.asarray(y)
    d.shape = (len(x), 1)
    np.save(name + '_annotations.npy', d)
    save_list_remove = glob.glob(name + '_annotations_*_out_of_*.npy')
    if save_list_remove:
        os.remove(save_list_remove[0])
    print('You are DONE!!!')
    return d


def wsi_cell_extraction_from_tiles(wsi_tiles_filename, im_size, coords, save_filename):
    tiles = np.load(wsi_tiles_filename)
    cells = []
    counter = 0
    num_of_columns = int(re.search('rows_(.+?)columns.npy', wsi_tiles_filename).group(1))
    num_of_rows = int(re.search('tiles_(.+?)rows', wsi_tiles_filename).group(1))

    for i in range(len(tiles)):
        # get coords of image
        column = i % num_of_columns
        row = i // num_of_columns
        # pixel range
        c_start = im_size * column
        c_stop = c_start + im_size
        r_start = im_size * row
        r_stop = r_start + im_size
        for j in coords:
            if c_start <= j[0] <= c_stop and r_start <= j[1] <= r_stop:
                crop_r_start = int(round(j[1] - r_start) - 32)
                crop_r_stop = int(round(j[1] - r_start) + 32)
                crop_c_start = int(round(j[0] - c_start) - 32)
                crop_c_stop = int(round(j[0] - c_start) + 32)
                if crop_r_start < 0 or crop_c_start < 0 or crop_r_stop > 1024 or crop_c_stop > 1024:
                    counter = counter + 1
                    print(counter, "out of", len(coords), "slices", "not sliced", 'src_im', i)
                    continue
                else:
                    cells.append(tiles[i][crop_r_start:crop_r_stop, crop_c_start:crop_c_stop, :3])
                    counter = counter + 1
                    print(counter, "out of", len(coords), "slices", int(round(j[1] - r_start) - 32),
                          int(round(j[1] - r_start) + 32), int(round(j[0] - c_start) - 32),
                          int(round(j[0] - c_start) + 32), 'src_im', i)
            else:
                continue
    np.save(save_filename)
    return cells


def anc_params_from_tiles(tiles, coords, wsi_tiles_filename, bbox_size=64, tile_size=1024):
    anc_params = []
    num_of_columns = int(re.search('rows_(.+?)columns.npy', wsi_tiles_filename).group(1))
    print('Extracting bounding boxes.')
    for i in range(len(tiles)):
        tile_bbox_param = []
        # get coords of image
        column = i % num_of_columns
        row = i // num_of_columns
        # pixel range
        c_start = tile_size * column
        c_stop = c_start + tile_size
        r_start = tile_size * row
        r_stop = r_start + tile_size
        print(i + 1, 'out of', len(tiles), '---', c_start, c_stop, r_start, r_stop)

        for j in range(len(coords)):
            if c_start <= coords[j, 0] < c_stop and r_start <= coords[j, 1] < r_stop:
                x = int(bbox_size / 2)
                y0 = int(round(coords[j, 1] - r_start) - x)
                y1 = int(round(coords[j, 1] - r_start) + x)
                x0 = int(round(coords[j, 0] - c_start) - x)
                x1 = int(round(coords[j, 0] - c_start) + x)
                params = np.asarray([[y0, x0, y1, x1]])
                print('--', params, coords[j], (params < 1024).sum() + (0 <= params).sum() == 8)
                if (params < 1024).sum() + (0 <= params).sum() == 8:
                    tile_bbox_param.append(params)
        if len(tile_bbox_param) != 0:
            tile_bbox_param = np.concatenate(tile_bbox_param, axis=0)
            anc_params.append(tile_bbox_param)
        else:
            anc_params.append('None')
    return anc_params


def convert_anc_to_box(anc_params, boundingbox, cls_key='cls-c4', cls_mask_key='cls-c4-msk', reg_key='reg-c4',
                       reg_mask_key='reg-c4-msk'):
    """
    ***This code is meant for just 1 feature map*** Will not handle multiple feature maps.
    """
    cls = []
    reg = []
    cls_mask = []
    reg_mask = []
    print('Converting extracted anchors to box parameters.')
    for i in range(len(anc_params)):
        print(i + 1, 'out of', len(anc_params))
        box = boundingbox.convert_anc_to_box(anc_params[i], np.ones((anc_params[i].shape[0], 1)))
        cls.append(box[cls_key])
        cls_mask.append(box[cls_mask_key])
        reg.append(box[reg_key])
        reg_mask.append(box[reg_mask_key])
    cls = np.array(cls)
    cls.shape = (len(cls),
                 boundingbox.params['anchor_gsizes'][0][0],
                 boundingbox.params['anchor_gsizes'][0][0],
                 boundingbox.params['classes'])
    cls_mask = np.array(cls_mask)
    cls_mask.shape = (len(cls_mask),
                      boundingbox.params['anchor_gsizes'][0][0],
                      boundingbox.params['anchor_gsizes'][0][0],
                      boundingbox.params['classes'])
    reg = np.array(reg)
    reg.shape = (len(reg),
                 boundingbox.params['anchor_gsizes'][0][0],
                 boundingbox.params['anchor_gsizes'][0][0],
                 4)
    reg_mask = np.array(reg_mask)
    reg_mask.shape = (len(reg_mask),
                      boundingbox.params['anchor_gsizes'][0][0],
                      boundingbox.params['anchor_gsizes'][0][0],
                      4)
    return cls, reg, cls_mask, reg_mask


def per_sample_tile_normalization(sorted_tiles):
    images = []
    for i in range(len(sorted_tiles)):
        # print(i + 1, 'out of', len(sorted_tiles))
        sample = sorted_tiles[i]
        # TODO: consider normalizing channels individually
        image = (sample - np.mean(sample)) / np.std(sample)
        images.append(image)
    images = np.array(images)
    return images


def normalized_tiles_and_bbox_params_from_wsi_tiles(wsi_tiles_filename, coords, boundingbox, normalize=False,
                                                    bbox_size=64, tile_size=1024):
    tiles = np.load(wsi_tiles_filename)
    if tiles.shape[3] == 4:
        tiles = tiles[:, :, :, :-1]
    anc = anc_params_from_tiles(tiles=tiles, coords=coords, wsi_tiles_filename=wsi_tiles_filename, bbox_size=bbox_size,
                                tile_size=tile_size)
    sorted_tiles = []
    sorted_anc = []
    print('Removing tiles without bounding boxes.')
    for i in range(len(anc)):
        print(i + 1, 'out of', len(anc))
        if anc[i] != 'None':
            sorted_tiles.append(tiles[i])
            sorted_anc.append((anc[i]))
    sorted_tiles = np.array(sorted_tiles)
    tiles = None
    cls, reg, cls_mask, reg_mask = convert_anc_to_box(sorted_anc, boundingbox)
    if normalize:
        images = per_sample_tile_normalization(sorted_tiles)
    else:
        images = sorted_tiles
    return images, cls, reg, cls_mask, reg_mask


def anc_params_from_mosaics(mosaic_metadata, coords, bbox_size=64, tile_size=1024):
    # TODO: check if this function is used and implement use of tile_size
    anc_params = []
    for i in range(len(mosaic_metadata)):
        tile_bbox_param = []
        r_start, r_stop, c_start, c_stop = wsi_mosaic_start_stop_slices(i)
        for j in range(len(coords)):
            if c_start <= coords[j, 0] < c_stop and r_start <= coords[j, 1] < r_stop:
                x = int(bbox_size / 2)
                y0 = int(round(coords[j, 1] - r_start) - x)
                y1 = int(round(coords[j, 1] - r_start) + x)
                x0 = int(round(coords[j, 0] - c_start) - x)
                x1 = int(round(coords[j, 0] - c_start) + x)
                params = np.asarray([[y0, x0, y1, x1]])
                print('--', params, coords[j], (params < 1024).sum() + (0 <= params).sum() == 8)
                if (params < 1024).sum() + (0 <= params).sum() == 8:
                    tile_bbox_param.append(params)
        if len(tile_bbox_param) != 0:
            tile_bbox_param = np.concatenate(tile_bbox_param, axis=0)
            anc_params.append(tile_bbox_param)
        else:
            anc_params.append('None')
    return anc_params


def normalized_tiles_and_bbox_params_from_wsi_mosaic(wsi_tiles_filename, mosaic_metadata, coords, boundingbox,
                                                     normalize=False,
                                                     bbox_size=64, tile_size=1024):
    mosaic = np.load(mosaic_metadata)
    tiles = np.load(wsi_tiles_filename)
    anc = anc_params_from_mosaics(mosaic_metadata=mosaic, coords=coords, bbox_size=bbox_size, tile_size=tile_size)
    sorted_tiles = []
    sorted_anc = []
    print('Removing tiles without bounding boxes.')
    for i in range(len(anc)):
        print(i + 1, 'out of', len(anc))
        if anc[i] != 'None':
            sorted_tiles.append(tiles[mosaic[i]])
            sorted_anc.append((anc[i]))
    sorted_tiles = np.array(sorted_tiles)
    cls, reg, cls_mask, reg_mask = convert_anc_to_box(sorted_anc, boundingbox)
    if normalize:
        images = per_sample_tile_normalization(sorted_tiles)
    else:
        images = sorted_tiles
    return images, cls, reg, cls_mask, reg_mask


def convert_anc_to_box_v2_2d(anc_params, boundingbox):
    """
    ***This code is meant for multiple feature maps***
    """
    boxes = {}
    for h in boundingbox.params['inputs_shapes'].keys():
        boxes[h] = []
    print('Converting extracted anchors to box parameters.')
    for i in range(len(anc_params)):
        print(i + 1, 'out of', len(anc_params))
        box = boundingbox.convert_anc_to_box(anc_params[i], np.ones((anc_params[i].shape[0], 1)))
        for j in box.keys():
            boxes[j].append(box[j])
    for k in boxes.keys():
        # this will provide an array with 4 dims for a 2d image, will not work for 3d!
        boxes[k] = np.concatenate(boxes[k], axis=0)
        print(k, boxes[k].shape)
    return boxes


def normalized_tiles_and_bbox_params_from_wsi_mosaic_v2(wsi_tiles_filename, mosaic_metadata, coords, boundingbox,
                                                        normalize=False, bbox_size=64, tile_size=1024):
    mosaic = np.load(mosaic_metadata)
    tiles = np.load(wsi_tiles_filename)
    if tiles.shape[3] == 4:
        tiles = tiles[:, :, :, :-1]
    anc = anc_params_from_mosaics(mosaic_metadata=mosaic, coords=coords, bbox_size=bbox_size, tile_size=tile_size)
    sorted_tiles = []
    sorted_anc = []
    print('Removing tiles without bounding boxes.')
    for i in range(len(anc)):
        print(i + 1, 'out of', len(anc))
        if anc[i] != 'None':
            sorted_tiles.append(tiles[mosaic[i]])
            sorted_anc.append((anc[i]))
    sorted_tiles = np.array(sorted_tiles)
    tiles = None
    im_boxes = convert_anc_to_box_v2_2d(sorted_anc, boundingbox)
    if normalize:
        images = per_sample_tile_normalization(sorted_tiles)
    else:
        images = sorted_tiles
    return images, im_boxes


def normalized_tiles_and_bbox_params_from_wsi_tiles_v2(wsi_tiles_filename, coords, boundingbox, normalize=False,
                                                       bbox_size=64, tile_size=1024):
    tiles = np.load(wsi_tiles_filename)
    if tiles.shape[3] == 4:
        tiles = tiles[:, :, :, :-1]
    anc = anc_params_from_tiles(tiles=tiles, coords=coords, wsi_tiles_filename=wsi_tiles_filename, bbox_size=bbox_size,
                                tile_size=tile_size)
    sorted_tiles = []
    sorted_anc = []
    print('Removing tiles without bounding boxes.')
    for i in range(len(anc)):
        print(i + 1, 'out of', len(anc))
        if anc[i] != 'None':
            sorted_tiles.append(tiles[i])
            sorted_anc.append((anc[i]))
    sorted_tiles = np.array(sorted_tiles)
    tiles = None
    im_boxes = convert_anc_to_box_v2_2d(sorted_anc, boundingbox)
    if normalize:
        images = per_sample_tile_normalization(sorted_tiles)
    else:
        images = sorted_tiles
    return images, im_boxes


# code to extract nuclei from wsi
def threshold_label_segment(wsi_tile):
    threshold = skimage.filters.threshold_isodata(wsi_tile[:, :, 2])
    binary_mask = wsi_tile[:, :, 2] > threshold
    binary_mask = ndimage.morphology.binary_fill_holes(skimage.morphology.dilation(
        skimage.morphology.dilation(skimage.morphology.erosion(skimage.morphology.erosion(binary_mask)))))
    binary_mask = np.uint8(binary_mask)

    # sure background
    sure_bg = np.uint8(
        skimage.morphology.dilation(skimage.morphology.dilation(skimage.morphology.dilation(binary_mask))))

    # sure foreground
    dist_transform = ndimage.distance_transform_edt(binary_mask)
    sure_fg_threshold = skimage.filters.threshold_isodata(dist_transform)
    sure_fg = np.uint8(dist_transform > sure_fg_threshold)

    # unknown region
    unknown = np.subtract(sure_bg, sure_fg)

    # label mask
    labels, _ = ndimage.label(sure_fg)
    labels = labels + 1
    labels[unknown == 1] = 0

    # add empty channels for watershed function
    binary_mask_3c = np.stack([binary_mask, np.zeros_like(binary_mask), np.zeros_like(binary_mask)], axis=2)

    # watershed segmented mask
    segmented_mask = cv.watershed(binary_mask_3c, labels)

    return segmented_mask, binary_mask


def get_wsi_coords_of_cells_from_tiles(i, coords, wsi_tiles_filename, tile_size):
    num_of_columns = int(re.search('rows_(.+?)columns.npy', wsi_tiles_filename).group(1))
    column = i % num_of_columns
    row = i // num_of_columns
    c_start = tile_size * column
    r_start = tile_size * row
    wsi_coords = []
    for j in coords:
        # TODO: revision comment: looks like code is treating coords as yx instead of xy,
        #  could be breaking other code that uses this
        # TODO: need to correct this and other code to use xy coords
        wsi_coords.append([r_start + j[0], c_start + j[1]])
    if len(wsi_coords) == 0:
        return None
    else:
        return np.array(wsi_coords)


def extract_cells_using_mask(wsi_tiles_filename, get_wsi_ccords=False, tile_size=1024):
    # function to get wsi coords is providing yx instead xy
    # TODO: make sure this code works once WSI_coords funciton is fixed
    wsi_tiles = np.load(wsi_tiles_filename)
    cells = []
    if get_wsi_ccords:
        wsi_coords = []
    for i in range(len(wsi_tiles)):
        if np.sum(wsi_tiles[i, :, :, 2]) == 0:
            continue
        segmented_mask, binary_mask = threshold_label_segment(wsi_tiles[i])
        # com_coords provides yx coords, not xy
        # TODO: edit com_coords so they are in xy format
        com_coords = ndimage.center_of_mass(binary_mask, segmented_mask, range(2, np.max(segmented_mask) + 1))
        com_coords = np.array(com_coords)
        com_coords = com_coords[~np.any(np.isnan(com_coords), axis=1)].astype(np.int32)
        delete1 = np.nonzero(com_coords - 32 < 0)[0]
        delete2 = np.nonzero(com_coords + 32 >= tile_size)[0]
        delete = np.concatenate([delete1, delete2])
        trimmed_com_coords = np.delete(com_coords, delete, axis=0)
        for j in trimmed_com_coords:
            cells.append(wsi_tiles[i, j[0] - 32:j[0] + 32, j[1] - 32:j[1] + 32, :])
        if get_wsi_ccords:
            wsi_coords.append(get_wsi_coords_of_cells_from_tiles(i, trimmed_com_coords, wsi_tiles_filename, tile_size))
        print(i + 1, 'out of', len(wsi_tiles))  # might replace with progress bar
    if get_wsi_ccords:
        cells = np.stack(cells)
        wsi_coords = [x for x in wsi_coords if x is not None]
        wsi_coords = np.concatenate(wsi_coords, axis=0)
        if len(wsi_coords) != len(cells):
            print('Number of extracted cells and coordinates do not match!')
        return cells, wsi_coords[:, ::-1]
    else:
        return np.array(cells)


def annotate_threshold_coords_using_manual_coords(thresh_coords, manual_coords, radius=20):
    mt = KDTree(manual_coords)
    nn = mt.query_radius(thresh_coords, radius, count_only=True)
    annotations = np.where(nn > 0, 1, 0)
    return annotations


def local_to_global_coords_retinanet(local_coords, wsi_tiles_filename, tile_size):
    # local_coords is a list of arrays which should be in xy format
    num_of_columns = int(re.search('rows_(.+?)columns.npy', wsi_tiles_filename).group(1))
    global_coords = []
    for i in range(len(local_coords)):
        if len(local_coords[i]) == 0:
            continue
        # get coords of tile
        column = i % num_of_columns
        row = i // num_of_columns
        # pixel range
        c_start = tile_size * column
        r_start = tile_size * row
        for j in range(len(local_coords[i])):
            global_coords.append([c_start + local_coords[i][j, 0], r_start + local_coords[i][j, 1]])
    if len(global_coords) == 0:
        return None
    else:
        return np.array(global_coords)


def convert_mosaic_coords_to_wsi(coords, metadata, wsi_tiles_filename, tile_size=1024):
    """

    :param coords:
    :type coords:
    :param metadata:
    :type metadata: array
    :param wsi_tiles_filename:
    :type wsi_tiles_filename:
    :param tile_size:
    :type tile_size:
    :return:
    :rtype:
    """
    num_of_columns = int(re.search('rows_(.+?)columns.npy', wsi_tiles_filename).group(1))
    # coords must be in xy format
    grid_coords = coords // 1025
    tile_coords = coords - (grid_coords * 1025)
    tile_index = grid_coords[:, 1] * 10 + grid_coords[:, 0]
    for i in range(len(tile_index)):
        tile_index[i] = metadata[tile_index[i]]  # mosaic tile_index now represents wsi tile index
    columns = tile_index % num_of_columns  # (n,) array
    rows = tile_index // num_of_columns  # (n,) array
    c_start = tile_size * columns  # (n,) array
    r_start = tile_size * rows  # (n,) array
    tile_start_coords = np.stack([c_start, r_start], axis=1)  # tile_start_coords are top left coords for each tile
    if tile_coords.shape == tile_start_coords.shape:
        return tile_coords + tile_start_coords
    else:
        print('arrays do not match')


def get_retinanet_training_dictionary_from_mosaic(wsi_tiles_filename, coords, boundingbox, normalize=False,
                                                  bbox_size=64, tile_size=1024):
    tiles, boxes = normalized_tiles_and_bbox_params_from_wsi_tiles_v2(
        wsi_tiles_filename,
        coords,
        boundingbox,
        normalize=normalize,
        bbox_size=bbox_size,
        tile_size=tile_size,
    )
    boxes['dat'] = tiles
    for key in boxes.keys():
        boxes[key] = np.expand_dims(boxes[key], axis=1)
    return boxes


def randomize_and_segregate_dataset_retinanet_dictionary(dictionary, training_percent=0.7, validation_percent=0.15,
                                                         test_percent=0.15):
    """
    Shuffles and splits images and box anchors into 3 dataset dictionaries:
    training, validation, and test datasets

    Keyword arguments:
    images -- an array of images
    labels -- an array of labels
    """

    p = np.random.permutation(len(dictionary['dat']))
    test_number = int(len(dictionary['dat']) * test_percent)
    validation_number = int(len(dictionary['dat']) * validation_percent)
    test_indices = p[0:test_number]
    validation_indices = p[test_number:(test_number + validation_number)]
    training_indices = p[(test_number + validation_number):]
    test_dic = {key: dictionary[key][test_indices] for key in dictionary.keys()}
    validation_dic = {key: dictionary[key][validation_indices] for key in dictionary.keys()}
    training_dic = {key: dictionary[key][training_indices] for key in dictionary.keys()}
    return training_dic, validation_dic, test_dic


def retinanet_generator(data, batchsize=1, normalize=True):
    i = 0
    keys = data.keys()
    while True:
        if i == (len(data['dat']) // batchsize):
            i = 0
            p = np.random.permutation(len(data['dat']))
            for key in keys:
                data[key] = data[key][p]
        start = i * batchsize
        stop = start + batchsize
        xbatch = {}
        ybatch = {}
        for key in keys:
            if 'dat' in key:
                if normalize:
                    xbatch[key] = per_sample_tile_normalization(data[key][start:stop])
                else:
                    xbatch[key] = data[key][start:stop]
            elif 'msk' in key:
                xbatch[key] = data[key][start:stop]
            else:
                ybatch[key] = data[key][start:stop]
        i += 1
        yield xbatch, ybatch


def retinanet_prediction_generator(images, boundingbox):
    # TODO: Not sure what this is for or when it was used. Consider deleting or making it an actual generator.
    pred_dic = {'dat': per_sample_tile_normalization(np.expand_dims(images, axis=1))}
    for key in boundingbox.params['inputs_shapes'].keys():
        if 'msk' in key:
            pred_dic[key] = np.zeros(shape=(len(images),)+tuple(boundingbox.params['input_shapes'][key]))
    return pred_dic


def retinanet_validation_generator(validation_dict):
    val_dict = validation_dict.copy()
    val_dict['dat'] = per_sample_tile_normalization(validation_dict['dat'])
    return val_dict


def retinanet_prediction(images, model, boundingbox):
    # want to add path boolean and code load model if model is a filepath
    output = model.predict(retinanet_prediction_generator(images, boundingbox))
    output_dic = {name: pred for name, pred in zip(model.output_names, output)}
    return output_dic


def cpec_coords_from_anc(anc, wsi_tiles_filename, tile_size, half_res=False):
    # anc is a list(2: anchor coords and classes) of a list of arrays (1 array per image tile)
    # anc arrays are local coordinates for the tile
    # convert tile coords to WSI coords
    local_coords = []
    for j in range(len(anc[0])):
        if len(anc[0][j]) != 0:
            y_coords = (anc[0][j][:, 2] + anc[0][j][:, 0]) / 2
            x_coords = (anc[0][j][:, 3] + anc[0][j][:, 1]) / 2
            local_coords.append(np.stack([x_coords, y_coords], axis=1))
        else:
            local_coords.append(anc[0][j])
    wsi_coords = local_to_global_coords_retinanet(local_coords, wsi_tiles_filename, tile_size)
    if not half_res:
        return wsi_coords
    elif half_res:
        return (wsi_coords * 2).astype(int)
    else:
        print('half_res parameter must be a boolean!')
        return


def tile_sample_hdf5_generator(tile_stack_filename, sample_size=100, name=None):
    # TODO: modify to allow re functions to work when not within the file folder or when providing a full file path
    # TODO: add naming override functionality
    num_of_columns = int(re.search('rows_(.+?)columns.npy', tile_stack_filename).group(1))
    num_of_rows = int(re.search('_(.{1,4}?)rows', tile_stack_filename).group(1))
    if name:
        wsi_name = name
    else:
        wsi_name = re.search('(^.{5,13}?)_', tile_stack_filename).group(1)
    filename = wsi_name + '_tile_sample_'
    previous_metadata = glob.glob(filename + '*.hdf5')
    max_previous = max([int(re.search('sample_(.{1,2}?).hdf5', i).group(1)) for i in previous_metadata]+[0])
    full_filename = filename + str(max_previous+1) + '.hdf5'
    tile_stack = np.load(tile_stack_filename)
    if max_previous == 0:
        p = np.random.permutation(len(tile_stack))
        delete_p = []
        for i in range(len(p)):
            if np.amax(tile_stack[p[i], :, :, 2]) < 85:
                delete_p.append(i)
        p = np.delete(p, delete_p, axis=0)
    else:
        f_old = h5py.File(filename + '1.hdf5', 'r')
        p = f_old['full_randomized_tile_indices']
    print('Saving...')
    f = h5py.File(full_filename, 'w')
    dset1 = f.create_dataset('images', data=tile_stack[p[:sample_size*(max_previous+1)]])
    dset2 = f.create_dataset('rows-columns', data=np.array([num_of_rows, num_of_columns]))
    dset3 = f.create_dataset('tile_index', data=p[:sample_size*(max_previous+1)])
    dset4 = f.create_dataset('full_randomized_tile_indices', data=p)
    print('Saved as:', full_filename)
    return


def unet_generator(imgs, masks):
    i = 0
    names = list(imgs.keys())
    while True:
        if i == len(names):
            i = 0
            p = np.random.permutation(len(names))
            names = names[p]
        img = per_sample_tile_normalization(np.expand_dims(imgs[names[i]][:], axis=0))
        msk = np.expand_dims(masks[names[i]][:], axis=0)
        i += 1
        yield img, msk


def sliding_window_generator(img, batchsize=16, window_size=64):
    # img should be a 4D array
    # window_size must be even
    windows = skimage.util.view_as_windows(np.pad(img,
                                                  ((int(window_size/2), int(window_size/2)-1),
                                                   (int(window_size/2), int(window_size/2)-1),
                                                   (int(window_size/2), int(window_size/2)-1),
                                                   (0, 0)),
                                                  'constant',
                                                  constant_values=0),
                                           (window_size,window_size,window_size,3),
                                           step=1)
    counter = 0
    im_batch = []
    for i in range(windows.shape[0]):
        for j in range(windows.shape[1]):
            for k in range(windows.shape[2]):
                # should I replace 0, in slice with :?
                im_batch.append(windows[i, j, k, 0, ...])
                counter += 1
                if counter == batchsize:
                    yield np.stack(im_batch)
                    counter = 0
                    im_batch=[]


def count_num_objs(msk, threshold, display_im=True):
    import copy
    labeled_msk = skimage.measure.label(msk, return_num=True, connectivity=3)
    #print('Before cleanup:', labeled_msk[1], 'objects')
    obj_vol = np.unique(labeled_msk[0], return_counts=True)
    # Number of object at each found voxel size
    vol_prevalence = np.unique(obj_vol[1], return_counts=True)
    obj_to_drop = obj_vol[0][obj_vol[1]<threshold]
    #print('After cleanup:', labeled_msk[1] - len(obj_to_drop), 'objects')
    msk2 = copy.deepcopy(labeled_msk[0])
    for label in obj_to_drop:
        msk2[msk2==label] = 0
    msk2[msk2>0] = 1
    if display_im:
        #before cleanup
        print('Before cleanup:', labeled_msk[1], 'objects')
        plt.imshow(np.max(msk, axis=0))
        plt.show()
        #after cleanup
        print('After cleanup:', labeled_msk[1] - len(obj_to_drop), 'objects')
        plt.imshow(np.max(msk2, axis=0))
        plt.show()
    return msk2


def generate_3d_binary_mask(image_fn_list, channel_index=1, voxel_threshold=50, sigma=1, kernel=(5,7,7), save_tiff=False, save_npy=True, src_fldr=None, dst_fldr=None):
    """
    *** Requires dev version of scipy for background subtraction ***
    Generates binary masked based on a single channel. Applies gaussian blurring followed by background subtraction
    using the rolling ball algorithm (similar to imagej) and finally generates a binary image based on ostu
    thresholding.
    :param image:
    :type image:
    :param channel_index:
    :type channel_index:
    :param voxel_threshold:
    :type voxel_threshold:
    :return:
    :rtype:
    """
    from skimage import filters, restoration
    def check_input_number(image_fn):
        while True:
            try:
                user_input = int(input(f'{image_fn} has more than 3 channels! Specify the desired channel index.'))
            except ValueError:
                print("Not an integer! Try again.")
                continue
            else:
                return user_input
                break

    def create_mask(image_fn, src_fldr=src_fldr, channel_index=channel_index, voxel_threshold=voxel_threshold, sigma=sigma, kernel=kernel):
        if src_fldr:
            image = skimage.io.imread(src_fldr + image_fn)
        else:
            image = skimage.io.imread(image_fn)
        if image.shape[-1] != 3:
            blurred_image = filters.gaussian(image[..., check_input_number(image_fn)], sigma=sigma, preserve_range=True)
        else:
            blurred_image = filters.gaussian(image[..., channel_index], sigma=sigma, preserve_range=True)
        background = restoration.rolling_ball(blurred_image, kernel=restoration.ellipsoid_kernel(kernel, 0.1))
        bkgrd_sub = blurred_image-background
        ostu_thresh = filters.threshold_otsu(bkgrd_sub)
        binary_msk = bkgrd_sub>=ostu_thresh
        cleanedup = count_num_objs(binary_msk, threshold=voxel_threshold, display_im=False)
        return cleanedup

    for image_fn in image_fn_list:
        print(f'Working on {image_fn}', end=' ')
        mask = create_mask(image_fn, channel_index=channel_index, voxel_threshold=voxel_threshold, sigma=sigma, kernel=kernel)
        if save_npy:
            if dst_fldr:
                np.save(dst_fldr + image_fn[:-4] + '_mask.npy', mask)
            else:
                np.save(image_fn[:-4] + '_mask.npy', mask)
        if save_tiff:
            mask[mask==1] = 255
            if dst_fldr:
                skimage.io.imsave(dst_fldr + image_fn[:-4] + '_mask.tif', mask.astype('uint8'))
            else:
                skimage.io.imsave(image_fn[:-4] + '_mask.tif', mask.astype('uint8'))
        print('\r', f'Finished {image_fn}')
    print('Done!')
    return


def quarter_divider(image):
    if image.ndim == 3:
        image = np.expand_dims(image, axis=-1)
    dim_length = image.shape[-2]
    if dim_length % 2 == 0:
        half_length = dim_length//2
        new_im1 = image[:, :half_length, :half_length, :]
        new_im2 = image[:, :half_length, half_length:, :]
        new_im3 = image[:, half_length:, :half_length, :]
        new_im4 = image[:, half_length:, half_length:, :]
        return new_im1, new_im2, new_im3, new_im4
    else:
        print('Image can not be divided evenly!')
        return None


def batch_quarter_divider(batch_fn, dst_fldr=None, save_tif=False):
    for fn in batch_fn:
        f_name = os.path.basename(os.path.normpath(fn))
        img_name, f_ext = os.path.splitext(f_name)
        img_name_path, _ = os.path.splitext(fn)
        if f_ext == '.npy':
            images = quarter_divider(np.load(fn))
        elif f_ext == '.tif':
            images = quarter_divider(skimage.io.imread(fn))
        else:
            print(f'{img_name} is not a npy or tif file! Please provide either a npy or tif file.')
            continue
        for i in range(4):
            if dst_fldr:
                np.save(f'{dst_fldr}{img_name}_quarter{i+1}.npy', images[i])
                if save_tif:
                    skimage.io.imsave(f'{dst_fldr}{img_name}_quarter{i+1}.tif', images[i])
            else:
                np.save(f'{img_name_path}_quarter{i + 1}.npy', images[i])
                if save_tif:
                    skimage.io.imsave(f'{img_name_path}_quarter{i + 1}.tif', images[i])


def generate_border(mask, bool_output=False):
    """
    Generates 3 pixel wide border region from binary mask
    :param bool_output:
    :type bool_output:
    :param mask:
    :type mask:
    :return:
    :rtype:
    """
    if bool_output:
        border = skimage.morphology.binary_dilation(mask - skimage.morphology.binary_erosion(mask))
    else:
        border = skimage.morphology.binary_dilation(mask - skimage.morphology.binary_erosion(mask)).astype(mask.dtype)
    return border


def add_border_to_mask(mask):
    border = generate_border(mask, bool_output=True)
    new_mask = np.where(border, 2, mask)
    return new_mask


def np_data_generator(images, labels, batch_size=16):
    i = 0
    while True:
        if i == (len(images)//batch_size):
            i = 0
            p = np.random.permutation(len(images))
            images = images[p]
            labels = labels[p]
        start = i * batch_size
        stop = start + batch_size
        xbatch = per_sample_tile_normalization(images[start:stop])
        ybatch = labels[start:stop]
        i += 1
        yield xbatch, ybatch


def np_validation_generator(images, labels, batch_size=16):
    i = 0
    while True:
        if i == (len(images) // batch_size):
            i = 0
        start = i * batch_size
        stop = start + batch_size
        xbatch = per_sample_tile_normalization(images[start:stop])
        ybatch = labels[start:stop]
        i += 1
        yield xbatch, ybatch


def np_prediction_generator(filename, batch_size=16):
    im = np.load(filename)
    if len(im)%batch_size == 0:
        for i in range(len(im)//batch_size):
            start = i * batch_size
            stop = start + batch_size
            xbatch = per_sample_tile_normalization(im[start:stop])
            yield xbatch
    else:
        for i in range((len(im)//batch_size)+1):
            start = i * batch_size
            if i == (len(im)//batch_size):
                xbatch = per_sample_tile_normalization(im[start:])
                yield xbatch
            else:
                stop = start + batch_size
                xbatch = per_sample_tile_normalization(im[start:stop])
                yield xbatch


def aggregate_retinanet_training_dictionaries(dicts):
    aggregated_dictionary = {key: np.concatenate([dic[key] for dic in dicts], axis=0) for key in dicts[0].keys()}


from tensorflow import losses, optimizers, nn, keras
from tensorflow.keras import Input, Model, models, layers, metrics
import os
import skimage
from skimage import transform
import matplotlib.pyplot as plt
import h5py
from tensorflow.keras.utils import plot_model, model_to_dot
import numpy as np
import pylab


def generate_3D_VGG(L2_constant, dropout, learning_rate, kernel_initializer='glorot_uniform', model_summary=True):
    # --- Define kwargs dictionary
    kwargs = {
        'kernel_size': (3, 3, 3),
        'padding': 'same',
        'kernel_regularizer': keras.regularizers.l2(L2_constant),
        'kernel_initializer': kernel_initializer}

    # --- Define block components
    conv = lambda x, filters, strides: layers.Conv3D(filters=filters, strides=strides, **kwargs)(x)
    relu = lambda x: layers.LeakyReLU()(x)
    norm = lambda x: layers.BatchNormalization()(x)
    # --- Define stride-1, stride-2 blocks
    conv1 = lambda filters, x: relu(norm(conv(x, filters, strides=1)))
    conv2 = lambda filters, x: relu(norm(conv(x, filters, strides=(2, 2, 2))))

    # --- Architecture
    inputs = Input(shape=(64, 64, 64, 3))
    l1 = conv2(16, conv1(16, inputs))
    l2 = conv2(32, conv1(32, l1))
    l3 = conv2(64, conv1(64, l2))
    l4 = conv2(128, conv1(128, l3))
    f0 = layers.Flatten()(l4)
    h1 = relu(layers.Dense(128)(f0))
    h2 = relu(norm(h1))
    h2 = layers.Dropout(dropout)
    logits = layers.Dense(2)(h1)

    model = Model(inputs=inputs, outputs=logits)

    if model_summary:
        model.summary()

    # --- Define a categorical cross-entropy loss
    loss = losses.SparseCategoricalCrossentropy(from_logits=True)
    optimizer = optimizers.Adam(learning_rate=learning_rate)

    # --- Compile model
    model.compile(
        optimizer=optimizer,
        loss=loss,
        metrics=['accuracy'])

    return model


def test_hyperparameters(dropout, LR, L2_reg, batch_size, epochs):
    model = generate_3D_VGG(L2_reg, dropout, LR, model_summary=False)
    hist = model.fit(x1, y, epochs=epochs, batch_size=batch_size, shuffle=True, validation_data=(x1_test, y_test))
    return hist


def grid_search_optimizaiton(dropout, LR, L2_reg, batch_size, epochs):
    history = []
    parameters = []
    import itertools
    for i, j, k, l in itertools.product(batch_size, dropout, LR, L2_reg):
        hist = test_hyperparameters(j, k, l, i, epochs)
        history.append(hist.history)
        np.save('properly_shuffled_WSI_vgg_' + str(j) + '_' + str(k) + '_' + str(l) + '_' + str(i) + '.npy',
                hist.history)
        parameters.append([i, j, k, l])
        print('batch size', i, 'dropout', j, 'learning rate', k, 'L2_constant', l)
        plot_training(hist)
    legend_acc1 = []
    print('batch_size', 'initializer', 'learning_rate', 'L2_constant')
    for m in range(len(history)):
        plt.plot(history[m]['accuracy'])
        i, j, k, l = parameters[m]
        legend_acc1.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model train accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(legend_acc1, loc='best')
    plt.show()

    legend_acc2 = []
    for m in range(len(history)):
        plt.plot(history[m]['val_accuracy'])
        i, j, k, l = parameters[m]
        legend_acc2.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model val accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(legend_acc2, loc='best')
    plt.show()

    legend_acc3 = []
    for m in range(len(history)):
        plt.plot(history[m]['loss'])
        i, j, k, l = parameters[m]
        legend_acc3.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model train loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(legend_acc3, loc='best')
    plt.show()

    legend_acc4 = []
    for m in range(len(history)):
        plt.plot(history[m]['val_loss'])
        i, j, k, l = parameters[m]
        legend_acc4.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model val loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(legend_acc4, loc='best')
    plt.show()
    return history, parameters


def ai_comparison_graphs(history, parameters):
    legend_acc1 = []
    print('batch_size', 'initializer', 'learning_rate', 'L2_constant')
    for m in range(len(history)):
        plt.plot(history[m]['accuracy'])
        i, j, k, l = parameters[m]
        legend_acc1.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model train accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(legend_acc1, loc='best')
    plt.show()

    legend_acc2 = []
    for m in range(len(history)):
        plt.plot(history[m]['val_accuracy'])
        i, j, k, l = parameters[m]
        legend_acc2.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model val accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(legend_acc2, loc='best')
    plt.show()

    legend_acc3 = []
    for m in range(len(history)):
        plt.plot(history[m]['loss'])
        i, j, k, l = parameters[m]
        legend_acc3.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model train loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(legend_acc3, loc='best')
    plt.show()

    legend_acc4 = []
    for m in range(len(history)):
        plt.plot(history[m]['val_loss'])
        i, j, k, l = parameters[m]
        legend_acc4.append(str(i) + ',' + str(j) + ',' + str(k) + ',' + str(l))
    plt.title('Model val loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(legend_acc4, loc='best')
    plt.show()




import numpy as np
import pylab
import tensorflow as tf
from tensorflow import losses, optimizers, nn, keras
from tensorflow.keras import Input, Model, models, layers, metrics
import os
import skimage
from skimage import transform
import matplotlib.pyplot as plt
from tensorflow.keras.utils import plot_model, model_to_dot
import h5py
from tensorflow.keras.datasets.mnist import load_data
import math


def discriminator(n_cat):
    conv2d = lambda x, filters, strides: layers.Conv2D(filters, (4, 4), strides=strides, padding='same')(x)
    leakyrelu = lambda x: layers.LeakyReLU(alpha=0.2)(x)
    norm = lambda x: layers.BatchNormalization()(x)
    dense = lambda x, nodes: layers.Dense(nodes)(x)
    flatten = lambda x : layers.Flatten()(x)

    inputs = Input(shape=(64, 64, 3))
    l1 = leakyrelu(conv2d(inputs, 64, (2, 2)))
    l2 = norm(leakyrelu(conv2d(l1, 128, (2, 2))))
    l3 = norm(leakyrelu(conv2d(l2, 256, (2, 2))))
    l4 = norm(leakyrelu(conv2d(l3, 256, (1, 1))))
    l5 = norm(leakyrelu(conv2d(l4, 256, (1, 1))))
    l6 = norm(leakyrelu(dense(flatten(l5), 1024)))
    d_out = layers.Dense(1, activation='sigmoid')(l6)

    # not sure if i should include the continuous variables.
    # for now, I will not implement continuous variables to keep things simple.
    aux1 = leakyrelu(norm(dense(l6, 128)))
    q_out = dense(aux1, n_cat)
    d_model = Model(inputs=inputs, outputs=d_out)
    opt = optimizers.Adam(learning_rate=0.0002, beta_1=0.5)
    d_model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])

    q_model = Model(inputs=inputs, outputs=q_out)
    return d_model, q_model


def generator():
    relu = lambda x: layers.ReLU()(x)
    norm = lambda x: layers.BatchNormalization()(x)
    trans2d = lambda x, filters, strides: layers.Conv2DTranspose(filters, (4, 4), strides=strides, padding='same')(x)
    dense = lambda x, nodes: layers.Dense(nodes)(x)

    inputs = Input(shape=gen_input_size)
    l1 = norm(relu(dense(inputs, 1024)))
    l2 = norm(relu(dense(l1, 8 * 8 * 256)))
    l3 = layers.Reshape((8,8,256))(l2)
    l4 = norm(relu(trans2d(l3, 256, (1, 1))))
    l5 = norm(relu(trans2d(l4, 256, (1, 1))))
    l6 = norm(relu(trans2d(l5, 128, (2, 2))))
    l7 = norm(relu(trans2d(l6, 64, (2, 2))))
    g_out = layers.Conv2DTranspose(3, (4, 4), strides=(2, 2), padding='same', activation='sigmoid')(l7)

    model = Model(inputs=inputs, outputs=g_out)
    return model


def gan(g_model, d_model, q_model):
    d_model.trainable = False
    d_output = d_model(g_model.output)
    q_output = q_model(g_model.output)

    model = Model(inputs=g_model.input, outputs=[d_output, q_output])

    opt = optimizers.Adam(learning_rate=0.0002, beta_1=0.5)
    model.compile(loss=['binary_crossentropy', 'categorical_crossentropy'], optimizer=opt)
    return model


def load_real_samples():
    (trainX, _), (_, _) = load_data()
    X = np.expand_dims(trainX, axis=-1)
    X = X.astype('float32')
    X = X / 255.0
    return X


def generate_real_samples(dataset, n_samples):
    ix = np.random.randint(0, dataset.shape[0], n_samples)
    X = dataset[ix]
    y = np.ones((n_samples, 1))
    return X, y


def generate_latent_points(latent_dim, n_samples, n_cat):
    x_input = np.random.randn(latent_dim * n_samples)
    x_input = x_input.reshape(n_samples, latent_dim)
    cat_codes = np.random.randint(0, n_cat, n_samples)
    cat_codes = keras.utils.to_categorical(cat_codes, num_classes=n_cat)
    x_input = np.hstack((x_input, cat_codes))
    return [x_input, cat_codes]


def generate_fake_samples(g_model, latent_dim, n_samples, n_cat):
    x_input = generate_latent_points(latent_dim, n_samples, n_cat)
    X = g_model.predict(x_input)
    y = np.zeros((n_samples, 1))
    return X, y


def save_plot(examples, epoch, n=10):
    for i in range(n * n):
        plt.subplot(n, n, 1 + i)
        plt.axis('off')
        plt.imshow(examples[i, :, :, :])
    filename = 'batchnorm_generated_plot_e%03d.png' % (epoch + 1)
    plt.savefig(filename)
    plt.close()


def summarize_performance(epoch, g_model, d_model, dataset, latent_dim, n_samples=100):
    X_real, y_real = generate_real_samples(dataset, n_samples)
    _, acc_real = d_model.evaluate(X_real, y_real, verbose=0)
    x_fake, y_fake = generate_fake_samples(g_model, latent_dim, n_samples, n_cat)
    _, acc_fake = d_model.evaluate(x_fake, y_fake, verbose=0)
    print('>Accuracy real: %.0f%%, fake: %.0f%%' % (acc_real * 100, acc_fake * 100))
    save_plot(x_fake, epoch)
    filename = 'batchnorm_generator_model_%03d.h5' % (epoch + 1)
    g_model.save(filename)


def train(g_model, d_model, gan_model, dataset, latent_dim, n_epochs=100, n_batch=256):
    bat_per_epo = int(dataset.shape[0] / n_batch)
    half_batch = int(n_batch / 2)
    for i in range(n_epochs):
        for j in range(bat_per_epo):
            X_real, y_real = generate_real_samples(dataset, half_batch)
            X_fake, y_fake = generate_fake_samples(g_model, latent_dim, half_batch, n_cat)
            X, y = np.vstack((X_real, X_fake)), np.vstack((y_real, y_fake))
            d_loss, _ = d_model.train_on_batch(X, y)
            X_gan, cat_codes = generate_latent_points(latent_dim, n_batch, n_cat)
            y_gan = np.ones((n_batch, 1))
            _, g_loss, q_loss = gan_model.train_on_batch(X_gan, [y_gan, cat_codes])
            print('>%d, %d/%d, d=%.3f, g=%.3f, q=%.3f' % (i + 1, j + 1, bat_per_epo, d_loss, g_loss, q_loss))
        if (i + 1) % 10 == 0:
            summarize_performance(i, g_model, d_model, dataset, latent_dim)


def generate_latent_points2(latent_dim, n_cat, n_samples, digit):
    # generate points in the latent space
    z_latent = np.random.randn(latent_dim * n_samples)
    # reshape into a batch of inputs for the network
    z_latent = z_latent.reshape(n_samples, latent_dim)
    # define categorical codes
    cat_codes = np.asarray([digit for _ in range(n_samples)])
    # one hot encode
    cat_codes = keras.utils.to_categorical(cat_codes, num_classes=n_cat)
    # concatenate latent points and control codes
    z_input = np.hstack((z_latent, cat_codes))
    return [z_input, cat_codes]


# create and save a plot of generated images
def save_plot2(examples, n_examples):
    # plot images
    for i in range(n_examples):
        # define subplot
        plt.subplot(math.sqrt(n_examples), math.sqrt(n_examples), 1 + i)
        # turn off axis
        plt.axis('off')
        # plot raw pixel data
        plt.imshow(examples[i, :, :, 0], cmap='gray_r')
    plt.show()


n_cat = 2
latent_dim = 128
gen_input_size = latent_dim + n_cat
# create the discriminator
d_model, q_model = discriminator(n_cat=10)
# create the generator
g_model = generator()
# create the gan
gan_model = gan(g_model, d_model, q_model)
# load image data
dataset = load_real_samples()

train(g_model, d_model, gan_model, dataset, latent_dim)

digit = 9
n_samples2 = 25
inputs, _ = generate_latent_points2(latent_dim, n_cat, n_samples2, digit)
X = g_model.predict(inputs)
save_plot2(X, n_samples2)
